# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module of `Data` sub class to represent a pseudopotential single file in UPF format and related utilities."""
import json
import re

from upf_to_json import upf_to_json
from .singlefile import SinglefileData

__all__ = ('UpfData',)

REGEX_UPF_VERSION = re.compile(r"""
    \s*<UPF\s+version\s*="
    (?P<version>.*)">
    """, re.VERBOSE)

REGEX_ELEMENT_V1 = re.compile(r"""
    (?P<element_name>[a-zA-Z]{1,2})
    \s+
    Element
    """, re.VERBOSE)

REGEX_ELEMENT_V2 = re.compile(
    r"""
    \s*
    element\s*=\s*(?P<quote_symbol>['"])\s*
    (?P<element_name>[a-zA-Z]{1,2})\s*
    (?P=quote_symbol)
    """, re.VERBOSE
)


def get_pseudos_from_structure(structure, family_name):
    """Return a dictionary mapping each kind name of the structure to corresponding `UpfData` from given family.

    :param structure: a `StructureData`
    :param family_name: the name of a UPF family group
    :return: dictionary mapping each structure kind name onto `UpfData` of corresponding element
    :raise aiida.common.MultipleObjectsError: if more than one UPF for the same element is found in the group.
    :raise aiida.common.NotExistent: if no UPF for an element in the group is found in the group.
    """
    from aiida.common.exceptions import NotExistent, MultipleObjectsError

    pseudo_list = {}
    family_pseudos = {}
    family = UpfData.get_upf_group(family_name)

    for node in family.nodes:
        if isinstance(node, UpfData):
            if node.element in family_pseudos:
                raise MultipleObjectsError(
                    f'More than one UPF for element {node.element} found in family {family_name}'
                )
            family_pseudos[node.element] = node

    for kind in structure.kinds:
        try:
            pseudo_list[kind.name] = family_pseudos[kind.symbol]
        except KeyError:
            raise NotExistent(f'No UPF for element {kind.symbol} found in family {family_name}')

    return pseudo_list


def upload_upf_family(folder, group_label, group_description, stop_if_existing=True):
    """Upload a set of UPF files in a given group.

    :param folder: a path containing all UPF files to be added.
        Only files ending in .UPF (case-insensitive) are considered.
    :param group_label: the name of the group to create. If it exists and is non-empty, a UniquenessError is raised.
    :param group_description: string to be set as the group description. Overwrites previous descriptions.
    :param stop_if_existing: if True, check for the md5 of the files and, if the file already exists in the DB, raises a
        MultipleObjectsError. If False, simply adds the existing UPFData node to the group.
    """
    # pylint: disable=too-many-locals,too-many-branches
    import os

    from aiida import orm
    from aiida.common import AIIDA_LOGGER
    from aiida.common.exceptions import UniquenessError
    from aiida.common.files import md5_file

    if not os.path.isdir(folder):
        raise ValueError('folder must be a directory')

    # only files, and only those ending with .upf or .UPF;
    # go to the real file if it is a symlink
    filenames = [
        os.path.realpath(os.path.join(folder, i))
        for i in os.listdir(folder)
        if os.path.isfile(os.path.join(folder, i)) and i.lower().endswith('.upf')
    ]

    nfiles = len(filenames)

    automatic_user = orm.User.objects.get_default()
    group, group_created = orm.UpfFamily.objects.get_or_create(label=group_label, user=automatic_user)

    if group.user.email != automatic_user.email:
        raise UniquenessError(
            'There is already a UpfFamily group with label {}'
            ', but it belongs to user {}, therefore you '
            'cannot modify it'.format(group_label, group.user.email)
        )

    # Always update description, even if the group already existed
    group.description = group_description

    # NOTE: GROUP SAVED ONLY AFTER CHECKS OF UNICITY

    pseudo_and_created = []

    for filename in filenames:
        md5sum = md5_file(filename)
        builder = orm.QueryBuilder()
        builder.append(UpfData, filters={'attributes.md5': {'==': md5sum}})
        existing_upf = builder.first()

        if existing_upf is None:
            # return the upfdata instances, not stored
            pseudo, created = UpfData.get_or_create(filename, use_first=True, store_upf=False)
            # to check whether only one upf per element exists
            # NOTE: actually, created has the meaning of "to_be_created"
            pseudo_and_created.append((pseudo, created))
        else:
            if stop_if_existing:
                raise ValueError(f'A UPF with identical MD5 to  {filename} cannot be added with stop_if_existing')
            existing_upf = existing_upf[0]
            pseudo_and_created.append((existing_upf, False))

    # check whether pseudo are unique per element
    elements = [(i[0].element, i[0].md5sum) for i in pseudo_and_created]
    # If group already exists, check also that I am not inserting more than
    # once the same element
    if not group_created:
        for aiida_n in group.nodes:
            # Skip non-pseudos
            if not isinstance(aiida_n, UpfData):
                continue
            elements.append((aiida_n.element, aiida_n.md5sum))

    elements = set(elements)  # Discard elements with the same MD5, that would
    # not be stored twice
    elements_names = [e[0] for e in elements]

    if not len(elements_names) == len(set(elements_names)):
        duplicates = {x for x in elements_names if elements_names.count(x) > 1}
        duplicates_string = ', '.join(i for i in duplicates)
        raise UniquenessError(f'More than one UPF found for the elements: {duplicates_string}.')

        # At this point, save the group, if still unstored
    if group_created:
        group.store()

    # save the upf in the database, and add them to group
    for pseudo, created in pseudo_and_created:
        if created:
            pseudo.store()

            AIIDA_LOGGER.debug(f'New node {pseudo.uuid} created for file {pseudo.filename}')
        else:
            AIIDA_LOGGER.debug(f'Reusing node {pseudo.uuid} for file {pseudo.filename}')

    # Add elements to the group all togetehr
    group.add_nodes([pseudo for pseudo, created in pseudo_and_created])

    nuploaded = len([_ for _, created in pseudo_and_created if created])

    return nfiles, nuploaded


def parse_upf(fname, check_filename=True):
    """
    Try to get relevant information from the UPF. For the moment, only the
    element name. Note that even UPF v.2 cannot be parsed with the XML minidom!
    (e.g. due to the & characters in the human-readable section).

    If check_filename is True, raise a ParsingError exception if the filename
    does not start with the element name.
    """
    import os

    from aiida.common.exceptions import ParsingError
    from aiida.common import AIIDA_LOGGER
    from aiida.orm.nodes.data.structure import _valid_symbols

    parsed_data = {}

    try:
        upf_contents = fname.read()
        fname = fname.name
    except AttributeError:
        with open(fname, encoding='utf8') as handle:
            upf_contents = handle.read()

    match = REGEX_UPF_VERSION.search(upf_contents)
    if match:
        version = match.group('version')
        AIIDA_LOGGER.debug(f'Version found: {version} for file {fname}')
    else:
        AIIDA_LOGGER.debug(f'Assuming version 1 for file {fname}')
        version = '1'

    parsed_data['version'] = version
    try:
        version_major = int(version.partition('.')[0])
    except ValueError:
        # If the version string does not contain a dot, fallback
        # to version 1
        AIIDA_LOGGER.debug(f'Falling back to version 1 for file {fname} version string {version} unrecognized')
        version_major = 1

    element = None
    if version_major == 1:
        match = REGEX_ELEMENT_V1.search(upf_contents)
        if match:
            element = match.group('element_name')
    else:  # all versions > 1
        match = REGEX_ELEMENT_V2.search(upf_contents)
        if match:
            element = match.group('element_name')

    if element is None:
        raise ParsingError(f'Unable to find the element of UPF {fname}')
    element = element.capitalize()
    if element not in _valid_symbols:
        raise ParsingError(f'Unknown element symbol {element} for file {fname}')
    if check_filename:
        if not os.path.basename(fname).lower().startswith(element.lower()):
            raise ParsingError(
                'Filename {0} was recognized for element '
                '{1}, but the filename does not start '
                'with {1}'.format(fname, element)
            )

    parsed_data['element'] = element

    return parsed_data


class UpfData(SinglefileData):
    """`Data` sub class to represent a pseudopotential single file in UPF format."""

    def __init__(self, file=None, filename=None, source=None, **kwargs):
        """Create UpfData instance from pseudopotential file.

          :param file: filepath or filelike object of the UPF potential file to store.
            Hint: Pass io.BytesIO(b"my string") to construct directly from a string.
          :param filename: specify filename to use (defaults to name of provided file).
          :param source: Dictionary with information on source of the potential (see ".source" property).
          """
        # pylint: disable=redefined-builtin
        super().__init__(file, filename=filename, **kwargs)
        if source is not None:
            self.set_source(source)

    @classmethod
    def get_or_create(cls, filepath, use_first=False, store_upf=True):
        """Get the `UpfData` with the same md5 of the given file, or create it if it does not yet exist.

        :param filepath: an absolute filepath on disk
        :param use_first: if False (default), raise an exception if more than one potential is found.
            If it is True, instead, use the first available pseudopotential.
        :param store_upf: boolean, if false, the `UpfData` if created will not be stored.
        :return: tuple of `UpfData` and boolean indicating whether it was created.
        """
        import os
        from aiida.common.files import md5_file

        if not os.path.isabs(filepath):
            raise ValueError('filepath must be an absolute path')

        pseudos = cls.from_md5(md5_file(filepath))

        if not pseudos:
            instance = cls(file=filepath)
            if store_upf:
                instance.store()
            return (instance, True)

        if len(pseudos) > 1:
            if use_first:
                return (pseudos[0], False)

            raise ValueError(
                'More than one copy of a pseudopotential with the same MD5 has been found in the DB. pks={}'.format(
                    ','.join([str(i.pk) for i in pseudos])
                )
            )

        return (pseudos[0], False)

    def store(self, *args, **kwargs):  # pylint: disable=signature-differs
        """Store the node, reparsing the file so that the md5 and the element are correctly reset."""
        from aiida.common.exceptions import ParsingError
        from aiida.common.files import md5_from_filelike

        if self.is_stored:
            return self

        with self.open(mode='r') as handle:
            parsed_data = parse_upf(handle)

        # Open in binary mode which is required for generating the md5 checksum
        with self.open(mode='rb') as handle:
            md5 = md5_from_filelike(handle)

        try:
            element = parsed_data['element']
        except KeyError:
            raise ParsingError(f'Could not parse the element from the UPF file {self.filename}')

        self.set_attribute('element', str(element))
        self.set_attribute('md5', md5)

        return super().store(*args, **kwargs)

    @classmethod
    def from_md5(cls, md5):
        """Return a list of all `UpfData` that match the given md5 hash.

        .. note:: assumes hash of stored `UpfData` nodes is stored in the `md5` attribute

        :param md5: the file hash
        :return: list of existing `UpfData` nodes that have the same md5 hash
        """
        from aiida.orm.querybuilder import QueryBuilder
        builder = QueryBuilder()
        builder.append(cls, filters={'attributes.md5': {'==': md5}})
        return builder.all(flat=True)

    def set_file(self, file, filename=None):
        """Store the file in the repository and parse it to set the `element` and `md5` attributes.

        :param file: filepath or filelike object of the UPF potential file to store.
            Hint: Pass io.BytesIO(b"my string") to construct the file directly from a string.
        :param filename: specify filename to use (defaults to name of provided file).
        """
        # pylint: disable=redefined-builtin
        from aiida.common.exceptions import ParsingError
        from aiida.common.files import md5_file, md5_from_filelike

        parsed_data = parse_upf(file)

        try:
            md5sum = md5_file(file)
        except TypeError:
            md5sum = md5_from_filelike(file)

        try:
            element = parsed_data['element']
        except KeyError:
            raise ParsingError(f"No 'element' parsed in the UPF file {self.filename}; unable to store")

        super().set_file(file, filename=filename)

        self.set_attribute('element', str(element))
        self.set_attribute('md5', md5sum)

    def get_upf_family_names(self):
        """Get the list of all upf family names to which the pseudo belongs."""
        from aiida.orm import UpfFamily
        from aiida.orm import QueryBuilder

        query = QueryBuilder()
        query.append(UpfFamily, tag='group', project='label')
        query.append(UpfData, filters={'id': {'==': self.id}}, with_group='group')
        return query.all(flat=True)

    @property
    def element(self):
        """Return the element of the UPF pseudopotential.

        :return: the element
        """
        return self.get_attribute('element', None)

    @property
    def md5sum(self):
        """Return the md5 checksum of the UPF pseudopotential file.

        :return: the md5 checksum
        """
        return self.get_attribute('md5', None)

    def _validate(self):
        """Validate the UPF potential file stored for this node."""
        from aiida.common.exceptions import ValidationError
        from aiida.common.files import md5_from_filelike

        super()._validate()

        with self.open(mode='r') as handle:
            parsed_data = parse_upf(handle)

        # Open in binary mode which is required for generating the md5 checksum
        with self.open(mode='rb') as handle:
            md5 = md5_from_filelike(handle)

        try:
            element = parsed_data['element']
        except KeyError:
            raise ValidationError(f"No 'element' could be parsed in the UPF {self.filename}")

        try:
            attr_element = self.get_attribute('element')
        except AttributeError:
            raise ValidationError("attribute 'element' not set.")

        try:
            attr_md5 = self.get_attribute('md5')
        except AttributeError:
            raise ValidationError("attribute 'md5' not set.")

        if attr_element != element:
            raise ValidationError(f"Attribute 'element' says '{attr_element}' but '{element}' was parsed instead.")

        if attr_md5 != md5:
            raise ValidationError(f"Attribute 'md5' says '{attr_md5}' but '{md5}' was parsed instead.")

    def _prepare_upf(self, main_file_name=''):
        """
        Return UPF content.
        """
        # pylint: disable=unused-argument
        return_string = self.get_content()

        return return_string.encode('utf-8'), {}

    @classmethod
    def get_upf_group(cls, group_label):
        """Return the UPF family group with the given label.

        :param group_label: the family group label
        :return: the `Group` with the given label, if it exists
        """
        from aiida.orm import UpfFamily

        return UpfFamily.get(label=group_label)

    @classmethod
    def get_upf_groups(cls, filter_elements=None, user=None):
        """Return all names of groups of type UpfFamily, possibly with some filters.

        :param filter_elements: A string or a list of strings.
            If present, returns only the groups that contains one UPF for every element present in the list. The default
            is `None`, meaning that all families are returned.
        :param user: if None (default), return the groups for all users.
            If defined, it should be either a `User` instance or the user email.
        :return: list of `Group` entities of type UPF.
        """
        from aiida.orm import UpfFamily
        from aiida.orm import QueryBuilder
        from aiida.orm import User

        builder = QueryBuilder()
        builder.append(UpfFamily, tag='group', project='*')

        if user:
            builder.append(User, filters={'email': {'==': user}}, with_group='group')

        if isinstance(filter_elements, str):
            filter_elements = [filter_elements]

        if filter_elements is not None:
            builder.append(UpfData, filters={'attributes.element': {'in': filter_elements}}, with_group='group')

        builder.order_by({UpfFamily: {'id': 'asc'}})

        return builder.all(flat=True)

    # pylint: disable=unused-argument
    def _prepare_json(self, main_file_name=''):
        """
        Returns UPF PP in json format.
        """
        with self.open() as file_handle:
            upf_json = upf_to_json(file_handle.read(), fname=self.filename)
        return json.dumps(upf_json).encode('utf-8'), {}



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Common import functions for both database backend"""
import copy
from typing import List, Optional

from aiida.common import timezone
from aiida.common.folders import RepositoryFolder
from aiida.common.progress_reporter import get_progress_reporter, create_callback
from aiida.orm import Group, ImportGroup, Node, QueryBuilder
from aiida.orm.utils._repository import Repository
from aiida.tools.importexport.archive.readers import ArchiveReaderAbstract
from aiida.tools.importexport.common import exceptions
from aiida.tools.importexport.dbimport.utils import IMPORT_LOGGER

MAX_COMPUTERS = 100
MAX_GROUPS = 100


def _copy_node_repositories(*, uuids_to_create: List[str], reader: ArchiveReaderAbstract):
    """Copy repositories of new nodes from the archive to the AiiDa profile.

    :param uuids_to_create: the node UUIDs to copy
    :param reader: the archive reader

    """
    if not uuids_to_create:
        return
    IMPORT_LOGGER.debug('CREATING NEW NODE REPOSITORIES...')
    with get_progress_reporter()(total=1, desc='Creating new node repos') as progress:

        _callback = create_callback(progress)

        for import_entry_uuid, subfolder in zip(
            uuids_to_create, reader.iter_node_repos(uuids_to_create, callback=_callback)
        ):
            destdir = RepositoryFolder(section=Repository._section_name, uuid=import_entry_uuid)  # pylint: disable=protected-access
            # Replace the folder, possibly destroying existing previous folders, and move the files
            # (faster if we are on the same filesystem, and in any case the source is a SandboxFolder)
            destdir.replace_with_folder(subfolder.abspath, move=True, overwrite=True)


def _make_import_group(*, group: Optional[ImportGroup], node_pks: List[int]) -> ImportGroup:
    """Make an import group containing all imported nodes.

    :param group: Use an existing group
    :param node_pks: node pks to add to group

    """
    # So that we do not create empty groups
    if not node_pks:
        IMPORT_LOGGER.debug('No nodes to import, so no import group created')
        return group

    # If user specified a group, import all things into it
    if not group:
        # Get an unique name for the import group, based on the current (local) time
        basename = timezone.localtime(timezone.now()).strftime('%Y%m%d-%H%M%S')
        counter = 0
        group_label = basename

        while Group.objects.find(filters={'label': group_label}):
            counter += 1
            group_label = f'{basename}_{counter}'

            if counter == MAX_GROUPS:
                raise exceptions.ImportUniquenessError(
                    f"Overflow of import groups (more than {MAX_GROUPS} groups exists with basename '{basename}')"
                )
        group = ImportGroup(label=group_label).store()

    # Add all the nodes to the new group
    builder = QueryBuilder().append(Node, filters={'id': {'in': node_pks}})

    first = True
    nodes = []
    description = 'Creating import Group - Preprocessing'

    with get_progress_reporter()(total=len(node_pks), desc=description) as progress:
        for entry in builder.iterall():
            if first:
                progress.set_description_str('Creating import Group', refresh=False)
                first = False
            progress.update()
            nodes.append(entry[0])

        group.add_nodes(nodes)
        progress.set_description_str('Done (cleaning up)', refresh=True)

    return group


def _sanitize_extras(fields: dict) -> dict:
    """Remove unwanted extra keys.

    :param fields: the database fields for the entity

    """
    fields = copy.copy(fields)
    fields['extras'] = {key: value for key, value in fields['extras'].items() if not key.startswith('_aiida_')}
    if fields.get('node_type', '').endswith('code.Code.'):
        fields['extras'] = {key: value for key, value in fields['extras'].items() if not key == 'hidden'}
    return fields



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`CalcJob` implementation to add two numbers using bash for testing and demonstration purposes."""
import warnings

from aiida.common.warnings import AiidaDeprecationWarning
from aiida.calculations.arithmetic.add import ArithmeticAddCalculation  # pylint: disable=unused-import

warnings.warn(  # pylint: disable=no-member
    'The add module has moved to aiida.calculations.arithmetic.add. '
    'This path will be removed in`v2.0.0`.', AiidaDeprecationWarning
)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Implementation of `Scheduler` base class."""
import abc

from aiida.common import exceptions, log
from aiida.common.escaping import escape_for_bash
from aiida.common.lang import classproperty
from aiida.schedulers.datastructures import JobResource, JobTemplate

__all__ = ('Scheduler', 'SchedulerError', 'SchedulerParsingError')


class SchedulerError(exceptions.AiidaException):
    pass


class SchedulerParsingError(SchedulerError):
    pass


class Scheduler(metaclass=abc.ABCMeta):
    """Base class for a job scheduler."""

    _logger = log.AIIDA_LOGGER.getChild('scheduler')

    # A list of features
    # Features that should be defined in the plugins:
    # 'can_query_by_user': True if I can pass the 'user' argument to
    # get_joblist_command (and in this case, no 'jobs' should be given).
    # Otherwise, if False, a list of jobs is passed, and no 'user' is given.
    _features = {}

    # The class to be used for the job resource.
    _job_resource_class = None

    def __str__(self):
        return self.__class__.__name__

    @classmethod
    def preprocess_resources(cls, resources, default_mpiprocs_per_machine=None):
        """Pre process the resources.

        Add the `num_mpiprocs_per_machine` key to the `resources` if it is not already defined and it cannot be deduced
        from the `num_machines` and `tot_num_mpiprocs` being defined. The value is also not added if the job resource
        class of this scheduler does not accept the `num_mpiprocs_per_machine` keyword. Note that the changes are made
        in place to the `resources` argument passed.
        """
        num_machines = resources.get('num_machines', None)
        tot_num_mpiprocs = resources.get('tot_num_mpiprocs', None)
        num_mpiprocs_per_machine = resources.get('num_mpiprocs_per_machine', None)

        if (
            num_mpiprocs_per_machine is None and cls.job_resource_class.accepts_default_mpiprocs_per_machine()  # pylint: disable=no-member
            and (num_machines is None or tot_num_mpiprocs is None)
        ):
            resources['num_mpiprocs_per_machine'] = default_mpiprocs_per_machine

    @classmethod
    def validate_resources(cls, **resources):
        """Validate the resources against the job resource class of this scheduler.

        :param resources: keyword arguments to define the job resources
        :raises ValueError: if the resources are invalid or incomplete
        """
        cls._job_resource_class.validate_resources(**resources)

    def __init__(self):
        self._transport = None

        if not issubclass(self._job_resource_class, JobResource):
            raise RuntimeError('the class attribute `_job_resource_class` is not a subclass of `JobResource`.')

    @classmethod
    def get_valid_schedulers(cls):
        """Return all available scheduler plugins.

        .. deprecated:: 1.3.0

            Will be removed in `2.0.0`, use `aiida.plugins.entry_point.get_entry_point_names` instead
        """
        import warnings
        from aiida.common.warnings import AiidaDeprecationWarning
        from aiida.plugins.entry_point import get_entry_point_names
        message = 'method is deprecated, use `aiida.plugins.entry_point.get_entry_point_names` instead'
        warnings.warn(message, AiidaDeprecationWarning)  # pylint: disable=no-member
        return get_entry_point_names('aiida.schedulers')

    @classmethod
    def get_short_doc(cls):
        """Return the first non-empty line of the class docstring, if available."""
        # Remove empty lines
        docstring = cls.__doc__
        if not docstring:
            return 'No documentation available'

        doclines = [i for i in docstring.splitlines() if i.strip()]
        if doclines:
            return doclines[0].strip()

        return 'No documentation available'

    def get_feature(self, feature_name):
        try:
            return self._features[feature_name]
        except KeyError:
            raise NotImplementedError(f'Feature {feature_name} not implemented for this scheduler')

    @property
    def logger(self):
        """Return the internal logger."""
        try:
            return self._logger
        except AttributeError:
            raise exceptions.InternalError('No self._logger configured for {}!')

    @classproperty
    def job_resource_class(cls):  # pylint: disable=no-self-argument
        return cls._job_resource_class

    @classmethod
    def create_job_resource(cls, **kwargs):
        """Create a suitable job resource from the kwargs specified."""
        # pylint: disable=not-callable
        return cls._job_resource_class(**kwargs)

    def get_submit_script(self, job_tmpl):
        """Return the submit script as a string.

        :parameter job_tmpl: a `aiida.schedulers.datastrutures.JobTemplate` instance.

        The plugin returns something like

        #!/bin/bash <- this shebang line is configurable to some extent
        scheduler_dependent stuff to choose numnodes, numcores, walltime, ...
        prepend_computer [also from calcinfo, joined with the following?]
        prepend_code [from calcinfo]
        output of _get_script_main_content
        postpend_code
        postpend_computer
        """
        if not isinstance(job_tmpl, JobTemplate):
            raise exceptions.InternalError('job_tmpl should be of type JobTemplate')

        empty_line = ''

        # I fill the list with the lines, and finally join them and return
        script_lines = []

        if job_tmpl.shebang:
            script_lines.append(job_tmpl.shebang)
        elif job_tmpl.shebang == '':
            # Here I check whether the shebang was set explicitly as an empty line.
            # In such a case, the first line is empty, if that's what the user wants:
            script_lines.append(job_tmpl.shebang)
        elif job_tmpl.shebang is None:
            script_lines.append('#!/bin/bash')
        else:
            raise ValueError(f'Invalid shebang set: {job_tmpl.shebang}')
        script_lines.append(self._get_submit_script_header(job_tmpl))
        script_lines.append(empty_line)

        if job_tmpl.prepend_text:
            script_lines.append(job_tmpl.prepend_text)
            script_lines.append(empty_line)

        script_lines.append(self._get_run_line(job_tmpl.codes_info, job_tmpl.codes_run_mode))
        script_lines.append(empty_line)

        if job_tmpl.append_text:
            script_lines.append(job_tmpl.append_text)
            script_lines.append(empty_line)

        footer = self._get_submit_script_footer(job_tmpl)  # pylint: disable=assignment-from-none
        if footer:
            script_lines.append(footer)
            script_lines.append(empty_line)

        return '\n'.join(script_lines)

    @abc.abstractmethod
    def _get_submit_script_header(self, job_tmpl):
        """Return the submit script header, using the parameters from the job template.

        :param job_tmpl: a `JobTemplate` instance with relevant parameters set.
        """

    def _get_submit_script_footer(self, job_tmpl):
        """Return the submit script final part, using the parameters from the job template.

        :param job_tmpl: a `JobTemplate` instance with relevant parameters set.
        """
        # pylint: disable=no-self-use,unused-argument
        return None

    def _get_run_line(self, codes_info, codes_run_mode):
        """Return a string with the line to execute a specific code with specific arguments.

        :parameter codes_info: a list of `aiida.common.datastructures.CodeInfo` objects. Each contains the information
            needed to run the code. I.e. `cmdline_params`, `stdin_name`, `stdout_name`, `stderr_name`, `join_files`. See
            the documentation of `JobTemplate` and `CodeInfo`.
        :parameter codes_run_mode: instance of `aiida.common.datastructures.CodeRunMode` contains the information on how
            to launch the multiple codes.
        :return: string with format: [executable] [args] {[ < stdin ]} {[ < stdout ]} {[2>&1 | 2> stderr]}
        """
        from aiida.common.datastructures import CodeRunMode

        list_of_runlines = []

        for code_info in codes_info:
            command_to_exec_list = []
            for arg in code_info.cmdline_params:
                command_to_exec_list.append(escape_for_bash(arg))
            command_to_exec = ' '.join(command_to_exec_list)

            stdin_str = f'< {escape_for_bash(code_info.stdin_name)}' if code_info.stdin_name else ''
            stdout_str = f'> {escape_for_bash(code_info.stdout_name)}' if code_info.stdout_name else ''

            join_files = code_info.join_files
            if join_files:
                stderr_str = '2>&1'
            else:
                stderr_str = f'2> {escape_for_bash(code_info.stderr_name)}' if code_info.stderr_name else ''

            output_string = f'{command_to_exec} {stdin_str} {stdout_str} {stderr_str}'

            list_of_runlines.append(output_string)

        self.logger.debug(f'_get_run_line output: {list_of_runlines}')

        if codes_run_mode == CodeRunMode.PARALLEL:
            list_of_runlines.append('wait\n')
            return ' &\n\n'.join(list_of_runlines)

        if codes_run_mode == CodeRunMode.SERIAL:
            return '\n\n'.join(list_of_runlines)

        raise NotImplementedError('Unrecognized code run mode')

    @abc.abstractmethod
    def _get_joblist_command(self, jobs=None, user=None):
        """Return the command to get the most complete description possible of currently active jobs.

        .. note::

            Typically one can pass only either jobs or user, depending on the specific plugin. The choice can be done
            according to the value returned by `self.get_feature('can_query_by_user')`

        :param jobs: either None to get a list of all jobs in the machine, or a list of jobs.
        :param user: either None, or a string with the username (to show only jobs of the specific user).
        """

    def _get_detailed_job_info_command(self, job_id):
        """Return the command to run to get detailed information for a given job.

        This is typically called after the job has finished, to retrieve the most detailed information possible about
        the job. This is done because most schedulers just make finished jobs disappear from the `qstat` command, and
        instead sometimes it is useful to know some more detailed information about the job exit status, etc.

        :raises: :class:`aiida.common.exceptions.FeatureNotAvailable`
        """
        # pylint: disable=no-self-use,not-callable,unused-argument
        raise exceptions.FeatureNotAvailable('Cannot get detailed job info')

    def get_detailed_job_info(self, job_id):
        """Return the detailed job info.

        This will be a dictionary with the return value, stderr and stdout content returned by calling the command that
        is returned by `_get_detailed_job_info_command`.

        :param job_id: the job identifier
        :return: dictionary with `retval`, `stdout` and `stderr`.
        """
        command = self._get_detailed_job_info_command(job_id)  # pylint: disable=assignment-from-no-return
        with self.transport:
            retval, stdout, stderr = self.transport.exec_command_wait(command)

        detailed_job_info = {
            'retval': retval,
            'stdout': stdout,
            'stderr': stderr,
        }

        return detailed_job_info

    def get_detailed_jobinfo(self, jobid):
        """
        Return a string with the output of the detailed_jobinfo command.

        .. deprecated:: 1.1.0
            Will be removed in `v2.0.0`, use :meth:`aiida.schedulers.scheduler.Scheduler.get_detailed_job_info` instead.

        At the moment, the output text is just retrieved
        and stored for logging purposes, but no parsing is performed.

        :raises: :class:`aiida.common.exceptions.FeatureNotAvailable`
        """
        import warnings
        from aiida.common.warnings import AiidaDeprecationWarning
        warnings.warn('function is deprecated, use `get_detailed_job_info` instead', AiidaDeprecationWarning)  # pylint: disable=no-member

        command = self._get_detailed_job_info_command(job_id=jobid)  # pylint: disable=assignment-from-no-return
        with self.transport:
            retval, stdout, stderr = self.transport.exec_command_wait(command)

        return f"""Detailed jobinfo obtained with command '{command}'
Return Code: {retval}
-------------------------------------------------------------
stdout:
{stdout}
stderr:
{stderr}
"""

    @abc.abstractmethod
    def _parse_joblist_output(self, retval, stdout, stderr):
        """Parse the joblist output as returned by executing the command returned by `_get_joblist_command` method.

        :return: list of `JobInfo` objects, one of each job each with at least its default params implemented.
        """

    def get_jobs(self, jobs=None, user=None, as_dict=False):
        """Return the list of currently active jobs.

        .. note:: typically, only either jobs or user can be specified. See also comments in `_get_joblist_command`.

        :param list jobs: a list of jobs to check; only these are checked
        :param str user: a string with a user: only jobs of this user are checked
        :param list as_dict: if False (default), a list of JobInfo objects is returned. If True, a dictionary is
            returned, having as key the job_id and as value the JobInfo object.
        :return: list of active jobs
        """
        with self.transport:
            retval, stdout, stderr = self.transport.exec_command_wait(self._get_joblist_command(jobs=jobs, user=user))

        joblist = self._parse_joblist_output(retval, stdout, stderr)
        if as_dict:
            jobdict = {job.job_id: job for job in joblist}
            if None in jobdict:
                raise SchedulerError('Found at least one job without jobid')
            return jobdict

        return joblist

    @property
    def transport(self):
        """Return the transport set for this scheduler."""
        if self._transport is None:
            raise SchedulerError('Use the set_transport function to set the transport for the scheduler first.')

        return self._transport

    def set_transport(self, transport):
        """Set the transport to be used to query the machine or to submit scripts.

        This class assumes that the transport is open and active.
        """
        self._transport = transport

    @abc.abstractmethod
    def _get_submit_command(self, submit_script):
        """Return the string to execute to submit a given script.

        .. warning:: the `submit_script` should already have been bash-escaped

        :param submit_script: the path of the submit script relative to the working directory.
        :return: the string to execute to submit a given script.
        """

    @abc.abstractmethod
    def _parse_submit_output(self, retval, stdout, stderr):
        """Parse the output of the submit command returned by calling the `_get_submit_command` command.

        :return: a string with the job ID.
        """

    def submit_from_script(self, working_directory, submit_script):
        """Submit the submission script to the scheduler.

        :return: return a string with the job ID in a valid format to be used for querying.
        """
        self.transport.chdir(working_directory)
        result = self.transport.exec_command_wait(self._get_submit_command(escape_for_bash(submit_script)))
        return self._parse_submit_output(*result)

    def kill(self, jobid):
        """Kill a remote job and parse the return value of the scheduler to check if the command succeeded.

        ..note::

            On some schedulers, even if the command is accepted, it may take some seconds for the job to actually
            disappear from the queue.

        :param jobid: the job ID to be killed
        :return: True if everything seems ok, False otherwise.
        """
        retval, stdout, stderr = self.transport.exec_command_wait(self._get_kill_command(jobid))
        return self._parse_kill_output(retval, stdout, stderr)

    @abc.abstractmethod
    def _get_kill_command(self, jobid):
        """Return the command to kill the job with specified jobid."""

    @abc.abstractmethod
    def _parse_kill_output(self, retval, stdout, stderr):
        """Parse the output of the kill command.

        :return: True if everything seems ok, False otherwise.
        """

    def parse_output(self, detailed_job_info, stdout, stderr):
        """Parse the output of the scheduler.

        :param detailed_job_info: dictionary with the output returned by the `Scheduler.get_detailed_job_info` command.
            This should contain the keys `retval`, `stdout` and `stderr` corresponding to the return value, stdout and
            stderr returned by the accounting command executed for a specific job id.
        :param stdout: string with the output written by the scheduler to stdout
        :param stderr: string with the output written by the scheduler to stderr
        :return: None or an instance of `aiida.engine.processes.exit_code.ExitCode`
        """
        raise exceptions.FeatureNotAvailable(f'output parsing is not available for `{self.__class__.__name__}`')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=cyclic-import
"""Futures that can poll or receive broadcasted messages while waiting for a task to be completed."""
import tornado.gen

import plumpy
import kiwipy

__all__ = ('ProcessFuture',)


class ProcessFuture(plumpy.Future):
    """Future that waits for a process to complete using both polling and listening for broadcast events if possible."""

    _filtered = None

    def __init__(self, pk, loop=None, poll_interval=None, communicator=None):
        """Construct a future for a process node being finished.

        If a None poll_interval is supplied polling will not be used. If a communicator is supplied it will be used
        to listen for broadcast messages.

        :param pk: process pk
        :param loop: An event loop
        :param poll_interval: optional polling interval, if None, polling is not activated.
        :param communicator: optional communicator, if None, will not subscribe to broadcasts.
        """
        from aiida.orm import load_node
        from .process import ProcessState

        super().__init__()
        assert not (poll_interval is None and communicator is None), 'Must poll or have a communicator to use'

        node = load_node(pk=pk)

        if node.is_terminated:
            self.set_result(node)
        else:
            self._communicator = communicator
            self.add_done_callback(lambda _: self.cleanup())

            # Try setting up a filtered broadcast subscriber
            if self._communicator is not None:
                broadcast_filter = kiwipy.BroadcastFilter(lambda *args, **kwargs: self.set_result(node), sender=pk)
                for state in [ProcessState.FINISHED, ProcessState.KILLED, ProcessState.EXCEPTED]:
                    broadcast_filter.add_subject_filter(f'state_changed.*.{state.value}')
                self._broadcast_identifier = self._communicator.add_broadcast_subscriber(broadcast_filter)

            # Start polling
            if poll_interval is not None:
                loop.add_callback(self._poll_process, node, poll_interval)

    def cleanup(self):
        """Clean up the future by removing broadcast subscribers from the communicator if it still exists."""
        if self._communicator is not None:
            self._communicator.remove_broadcast_subscriber(self._broadcast_identifier)
            self._communicator = None
            self._broadcast_identifier = None

    @tornado.gen.coroutine
    def _poll_process(self, node, poll_interval):
        """Poll whether the process node has reached a terminal state."""
        while not self.done() and not node.is_terminated:
            yield tornado.gen.sleep(poll_interval)

        if not self.done():
            self.set_result(node)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member,no-name-in-module,import-error
"""This migration creates UUID column and populates it with distinct UUIDs

This migration corresponds to the 0024_dblog_update Django migration.

Revision ID: ea2f50e7f615
Revises: 041a79fc615f
Create Date: 2019-01-30 19:22:50.984380"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'ea2f50e7f615'
down_revision = '041a79fc615f'
branch_labels = None
depends_on = None


def set_new_uuid(connection):
    """
    Set new and distinct UUIDs to all the logs
    """
    from aiida.common.utils import get_new_uuid

    # Exit if there are no rows - e.g. initial setup
    id_query = connection.execute('SELECT db_dblog.id FROM db_dblog')
    if id_query.rowcount == 0:
        return

    id_res = id_query.fetchall()
    ids = list()
    for (curr_id,) in id_res:
        ids.append(curr_id)
    uuids = set()
    while len(uuids) < len(ids):
        uuids.add(get_new_uuid())

    # Create the key/value pairs
    key_values = ','.join("({}, '{}')".format(curr_id, curr_uuid) for curr_id, curr_uuid in zip(ids, uuids))

    update_stm = f"""
        UPDATE db_dblog as t SET
            uuid = uuid(c.uuid)
        from (values {key_values}) as c(id, uuid) where c.id = t.id"""
    connection.execute(update_stm)


def upgrade():
    """ Add an UUID column an populate it with unique UUIDs """
    from aiida.common.utils import get_new_uuid
    connection = op.get_bind()

    # Create the UUID column
    op.add_column(
        'db_dblog', sa.Column('uuid', postgresql.UUID(), autoincrement=False, nullable=True, default=get_new_uuid)
    )

    # Populate the uuid column
    set_new_uuid(connection)


def downgrade():
    """ Remove the UUID column """
    op.drop_column('db_dblog', 'uuid')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""General utilities for Transport classes."""

import time

from paramiko import ProxyCommand

from aiida.common.extendeddicts import FixedFieldsAttributeDict


class FileAttribute(FixedFieldsAttributeDict):
    """
    A class, resembling a dictionary, to describe the attributes of a file,
    that is returned by get_attribute().
    Possible keys: st_size, st_uid, st_gid, st_mode, st_atime, st_mtime
    """
    _valid_fields = (
        'st_size',
        'st_uid',
        'st_gid',
        'st_mode',
        'st_atime',
        'st_mtime',
    )


class _DetachedProxyCommand(ProxyCommand):
    """Modifies paramiko's ProxyCommand by launching the process in a separate process group."""

    def __init__(self, command_line):  # pylint: disable=super-init-not-called
        # Note that the super().__init__ MUST NOT be called here, otherwise two subprocesses will be created.
        from subprocess import Popen, PIPE
        from shlex import split as shlsplit

        self.cmd = shlsplit(command_line)
        self.process = Popen(self.cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, bufsize=0, start_new_session=True)
        self.timeout = None

    def close(self):
        try:
            self.process.terminate()
        # In case the process doesn't exist anymore
        except OSError:
            pass
        for _ in range(10):
            if self.process.poll() is not None:
                break
            time.sleep(0.2)
        else:
            try:
                self.process.kill()
            # In case the process doesn't exist anymore
            except OSError:
                pass
            for _ in range(10):
                if self.process.poll() is not None:
                    break
                time.sleep(0.2)


def copy_from_remote_to_remote(transportsource, transportdestination, remotesource, remotedestination, **kwargs):
    """
    Copy files or folders from a remote computer to another remote computer.

    :param transportsource: transport to be used for the source computer
    :param transportdestination: transport to be used for the destination computer
    :param str remotesource: path to the remote source directory / file
    :param str remotedestination: path to the remote destination directory / file
    :param kwargs: keyword parameters passed to the final put,
        except for 'dereference' that is passed to the initial get

    .. note:: it uses the method transportsource.copy_from_remote_to_remote
    """
    transportsource.copy_from_remote_to_remote(transportdestination, remotesource, remotedestination, **kwargs)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Archive writer classes."""
from abc import ABC, abstractmethod
from copy import deepcopy
import os
from pathlib import Path
import shelve
import shutil
import time
import tempfile
from types import TracebackType
from typing import Any, cast, Dict, List, Optional, Type, Union
import zipfile

from archive_path import TarPath, ZipPath

from aiida.common import json
from aiida.common.exceptions import InvalidOperation
from aiida.common.folders import Folder
from aiida.tools.importexport.archive.common import ArchiveMetadata
from aiida.tools.importexport.common.config import (
    EXPORT_VERSION, NODE_ENTITY_NAME, NODES_EXPORT_SUBFOLDER, ExportFileFormat
)
from aiida.tools.importexport.common.utils import export_shard_uuid

__all__ = ('ArchiveWriterAbstract', 'get_writer', 'WriterJsonZip', 'WriterJsonTar', 'WriterJsonFolder')


def get_writer(file_format: str) -> Type['ArchiveWriterAbstract']:
    """Return the available writer classes."""
    writers = {
        ExportFileFormat.ZIP: WriterJsonZip,
        ExportFileFormat.TAR_GZIPPED: WriterJsonTar,
        'folder': WriterJsonFolder,
        'null': WriterNull,
    }

    if file_format not in writers:
        raise ValueError(
            f'Can only write in the formats: {tuple(writers.keys())}, not {file_format}, '
            'please specify one for "file_format".'
        )

    return cast(Type[ArchiveWriterAbstract], writers[file_format])


class ArchiveWriterAbstract(ABC):
    """An abstract interface for AiiDA archive writers."""

    def __init__(self, filepath: Union[str, Path], **kwargs: Any):
        """Initiate the writer.

        :param filepath: the path to the file to export to.
        :param kwargs: keyword arguments specific to the writer implementation.

        """
        # pylint: disable=unused-argument
        self._filepath = Path(filepath)
        self._info: Dict[str, Any] = {}
        self._in_context: bool = False

    @property
    def filepath(self) -> Path:
        """Return the filepath to write to."""
        return self._filepath

    @property
    def export_info(self) -> Dict[str, Any]:
        """Return information stored during the export process."""
        return deepcopy(self._info)

    @property
    @abstractmethod
    def file_format_verbose(self) -> str:
        """Return the file format name."""

    @property
    @abstractmethod
    def export_version(self) -> str:
        """Return the export version."""

    def __enter__(self) -> 'ArchiveWriterAbstract':
        """Open the contextmanager """
        self._in_context = True
        # reset the export information
        self._info = {'writer_entered': time.time()}
        self.open()
        return self

    def __exit__(
        self, exctype: Optional[Type[BaseException]], excinst: Optional[BaseException], exctb: Optional[TracebackType]
    ):
        """Open the contextmanager """
        self.close(exctype is not None)
        self._in_context = False
        self._info['writer_exited'] = time.time()
        return False

    def assert_within_context(self):
        """Assert that the method is called within a context.

        :raises: `~aiida.common.exceptions.InvalidOperation`: if not called within a context
        """
        if not self._in_context:
            raise InvalidOperation('the ArchiveReader method should be used within a context')

    def add_export_info(self, key: str, value: Any):
        """Record information about the export process.

        This information can be accessed by ``writer.export_info``,
        it is reset on entrance of the context manager.

        """
        self._info[key] = value

    def _remove_filepath(self):
        """To run before moving the final export from a temporary folder."""
        if self.filepath.exists():
            if self.filepath.is_file():
                self.filepath.unlink()
            else:
                shutil.rmtree(str(self.filepath))

    # initialise/finalise methods

    @abstractmethod
    def open(self):
        """Setup the export process.

        This method is called on entering a context manager.

        """

    @abstractmethod
    def close(self, excepted: bool):
        """Finalise the export.

        This method is called on exiting a context manager.

        :param excepted: Whether

        """

    # write methods

    @abstractmethod
    def write_metadata(self, data: ArchiveMetadata):
        """Write the metadata of the export process."""

    @abstractmethod
    def write_link(self, data: Dict[str, str]):
        """Write a dictionary of information for a single provenance link.

        :param data: ``{'input': <UUID_STR>, 'output': <UUID_STR>, 'label': <LABEL_STR>, 'type': <TYPE_STR>}``

        """

    @abstractmethod
    def write_group_nodes(self, uuid: str, node_uuids: List[str]):
        """Write a mapping of a group to the nodes it contains.

        :param uuid: the UUID of the group
        :param node_uuids: the list of node UUIDs the group contains

        """

    @abstractmethod
    def write_entity_data(self, name: str, pk: int, id_key: str, fields: Dict[str, Any]):
        """Write the data for a single DB entity.

        :param name: the name of the entity (e.g. 'NODE')
        :param pk: the primary key of the entity (unique for the current DB only)
        :param id_key: the key within ``fields`` that denotes the unique identifier of the entity (unique for all DBs)
        :param fields: mapping of DB fields to store -> values

        """

    @abstractmethod
    def write_node_repo_folder(self, uuid: str, path: Union[str, Path], overwrite: bool = True):
        """Write a node repository to the archive.

        :param uuid: The UUID of the node
        :param path: The path to the repository folder on disk
        :param overwrite: Allow to overwrite existing path in archive

        """


class WriterNull(ArchiveWriterAbstract):
    """A null archive writer, which does not do anything."""

    @property
    def file_format_verbose(self) -> str:
        return 'Null'

    @property
    def export_version(self) -> str:
        return EXPORT_VERSION

    def open(self):
        pass

    def close(self, excepted: bool):
        pass

    def write_metadata(self, data: ArchiveMetadata):
        pass

    def write_link(self, data: Dict[str, str]):
        pass

    def write_group_nodes(self, uuid: str, node_uuids: List[str]):
        pass

    def write_entity_data(self, name: str, pk: int, id_key: str, fields: Dict[str, Any]):
        pass

    def write_node_repo_folder(self, uuid: str, path: Union[str, Path], overwrite: bool = True):
        pass


class WriterJsonZip(ArchiveWriterAbstract):
    """An archive writer,
    which writes database data as a single JSON and repository data in a zipped folder system.
    """

    def __init__(
        self, filepath: Union[str, Path], *, use_compression: bool = True, cache_zipinfo: bool = False, **kwargs
    ):
        """Initiate the writer.

        :param filepath: the path to the file to export to.
        :param use_compression: Whether or not to deflate the objects inside the zip file.
        :param cache_zipinfo: Cache the zip file index on disk during the write.
            This reduces the RAM usage of the process, but will make the process slower.

        """
        super().__init__(filepath)
        self._compression = zipfile.ZIP_DEFLATED if use_compression else zipfile.ZIP_STORED
        self._cache_zipinfo = cache_zipinfo

    @property
    def file_format_verbose(self) -> str:
        return f'JSON Zip (compression={self._compression})'

    @property
    def export_version(self) -> str:
        return EXPORT_VERSION

    def open(self):
        # pylint: disable=attribute-defined-outside-init
        self.assert_within_context()
        # create a temporary folder in which to perform the write
        self._temp_path: Path = Path(tempfile.mkdtemp())
        # open a zipfile in in write mode to export to
        self._zipinfo_cache: Optional[dict]
        if self._cache_zipinfo:
            self._zipinfo_cache = shelve.open(str(self._temp_path / 'zipinfo_cache'))  # type: ignore
        else:
            self._zipinfo_cache = None
        self._archivepath: ZipPath = ZipPath(
            self._temp_path / 'export', mode='w', compression=self._compression, name_to_info=self._zipinfo_cache
        )
        # setup data to store
        self._data: Dict[str, Any] = {
            'node_attributes': {},
            'node_extras': {},
            'export_data': {},
            'links_uuid': [],
            'groups_uuid': {},
        }

    def close(self, excepted: bool):
        self.assert_within_context()
        if excepted:
            self._archivepath.close()
            shutil.rmtree(self._temp_path)
            return
        # write data.json
        with self._archivepath.joinpath('data.json').open('wb') as handle:
            json.dump(self._data, handle)
        # close the zipfile to finalise write
        self._archivepath.close()
        if getattr(self, '_zipinfo_cache', None) is not None:
            self._zipinfo_cache.close()  # type: ignore
            delattr(self, '_zipinfo_cache')
        # move the compressed file to the final path
        self._remove_filepath()
        shutil.move(str(self._archivepath.filepath), str(self.filepath))
        # remove the temporary folder
        shutil.rmtree(self._temp_path)

    def write_metadata(self, data: ArchiveMetadata):
        metadata = {
            'export_version': self.export_version,
            'aiida_version': data.aiida_version,
            'all_fields_info': data.all_fields_info,
            'unique_identifiers': data.unique_identifiers,
            'export_parameters': {
                'graph_traversal_rules': data.graph_traversal_rules,
                'entities_starting_set': data.entities_starting_set,
                'include_comments': data.include_comments,
                'include_logs': data.include_logs,
            },
            'conversion_info': data.conversion_info
        }
        with self._archivepath.joinpath('metadata.json').open('wb') as handle:
            json.dump(metadata, handle)

    def write_link(self, data: Dict[str, str]):
        self._data['links_uuid'].append(data)

    def write_group_nodes(self, uuid: str, node_uuids: List[str]):
        self._data['groups_uuid'][uuid] = node_uuids

    def write_entity_data(self, name: str, pk: int, id_key: str, fields: Dict[str, Any]):
        if name == NODE_ENTITY_NAME:
            # perform translation to current internal format
            self._data['node_attributes'][pk] = fields.pop('attributes')
            self._data['node_extras'][pk] = fields.pop('extras')
        self._data['export_data'].setdefault(name, {})[pk] = fields

    def write_node_repo_folder(self, uuid: str, path: Union[str, Path], overwrite: bool = True):
        self.assert_within_context()
        (self._archivepath / NODES_EXPORT_SUBFOLDER / export_shard_uuid(uuid)).puttree(path, check_exists=not overwrite)


class WriterJsonTar(ArchiveWriterAbstract):
    """An archive writer,
    which writes database data as a single JSON and repository data in a folder system.

    The entire containing folder is then compressed as a tar file.
    """

    @property
    def file_format_verbose(self) -> str:
        return 'Gzipped tarball (compressed)'

    @property
    def export_version(self) -> str:
        return EXPORT_VERSION

    def open(self):
        # pylint: disable=attribute-defined-outside-init
        self.assert_within_context()
        # create a temporary folder in which to perform the write
        self._temp_path: Path = Path(tempfile.mkdtemp())
        # open a zipfile in in write mode to export to
        self._archivepath: TarPath = TarPath(self._temp_path / 'export', mode='w:gz', dereference=True)
        # setup data to store
        self._data: Dict[str, Any] = {
            'node_attributes': {},
            'node_extras': {},
            'export_data': {},
            'links_uuid': [],
            'groups_uuid': {},
        }

    def close(self, excepted: bool):
        self.assert_within_context()
        if excepted:
            self._archivepath.close()
            shutil.rmtree(self._temp_path)
            return
        # write data.json
        with self._archivepath.joinpath('data.json').open('wb') as handle:
            json.dump(self._data, handle)
        # compress
        # close the zipfile to finalise write
        self._archivepath.close()
        # move the compressed file to the final path
        self._remove_filepath()
        shutil.move(str(self._archivepath.filepath), str(self.filepath))
        # remove the temporary folder
        shutil.rmtree(self._temp_path)

    def write_metadata(self, data: ArchiveMetadata):
        metadata = {
            'export_version': self.export_version,
            'aiida_version': data.aiida_version,
            'all_fields_info': data.all_fields_info,
            'unique_identifiers': data.unique_identifiers,
            'export_parameters': {
                'graph_traversal_rules': data.graph_traversal_rules,
                'entities_starting_set': data.entities_starting_set,
                'include_comments': data.include_comments,
                'include_logs': data.include_logs,
            },
            'conversion_info': data.conversion_info
        }
        with self._archivepath.joinpath('metadata.json').open('wb') as handle:
            json.dump(metadata, handle)

    def write_link(self, data: Dict[str, str]):
        self._data['links_uuid'].append(data)

    def write_group_nodes(self, uuid: str, node_uuids: List[str]):
        self._data['groups_uuid'][uuid] = node_uuids

    def write_entity_data(self, name: str, pk: int, id_key: str, fields: Dict[str, Any]):
        if name == NODE_ENTITY_NAME:
            # perform translation to current internal format
            self._data['node_attributes'][pk] = fields.pop('attributes')
            self._data['node_extras'][pk] = fields.pop('extras')
        self._data['export_data'].setdefault(name, {})[pk] = fields

    def write_node_repo_folder(self, uuid: str, path: Union[str, Path], overwrite: bool = True):
        self.assert_within_context()
        (self._archivepath / NODES_EXPORT_SUBFOLDER / export_shard_uuid(uuid)).puttree(path, check_exists=not overwrite)


class WriterJsonFolder(ArchiveWriterAbstract):
    """An archive writer,
    which writes database data as a single JSON and repository data in a folder system.

    This writer is mainly intended for backward compatibility with `export_tree`.
    """

    def __init__(self, filepath: str, folder: Folder = None, **kwargs: Any):
        """Initiate the writer.

        :param folder: a folder to write the archive to.
        :param filepath: dummy value not used

        """
        super().__init__(filepath, **kwargs)
        if not isinstance(folder, Folder):
            raise TypeError('`folder` must be specified and given as an AiiDA Folder entity')
        self._folder = cast(Folder, folder)

    @property
    def file_format_verbose(self) -> str:
        return 'JSON Folder'

    @property
    def export_version(self) -> str:
        return EXPORT_VERSION

    def open(self):
        # pylint: disable=attribute-defined-outside-init
        self.assert_within_context()
        # setup data to store
        self._data: Dict[str, Any] = {
            'node_attributes': {},
            'node_extras': {},
            'export_data': {},
            'links_uuid': [],
            'groups_uuid': {},
        }
        # ensure folder is created
        self._folder.create()

    def close(self, excepted: bool):
        self.assert_within_context()
        if excepted:
            return
        with self._folder.open('data.json', 'wb') as handle:
            json.dump(self._data, handle)

    def write_metadata(self, data: ArchiveMetadata):
        metadata = {
            'export_version': self.export_version,
            'aiida_version': data.aiida_version,
            'all_fields_info': data.all_fields_info,
            'unique_identifiers': data.unique_identifiers,
            'export_parameters': {
                'graph_traversal_rules': data.graph_traversal_rules,
                'entities_starting_set': data.entities_starting_set,
                'include_comments': data.include_comments,
                'include_logs': data.include_logs,
            },
            'conversion_info': data.conversion_info
        }
        with self._folder.open('metadata.json', 'wb') as handle:
            json.dump(metadata, handle)

    def write_link(self, data: Dict[str, str]):
        self._data['links_uuid'].append(data)

    def write_group_nodes(self, uuid: str, node_uuids: List[str]):
        self._data['groups_uuid'][uuid] = node_uuids

    def write_entity_data(self, name: str, pk: int, id_key: str, fields: Dict[str, Any]):
        if name == NODE_ENTITY_NAME:
            # perform translation to current internal format
            self._data['node_attributes'][pk] = fields.pop('attributes')
            self._data['node_extras'][pk] = fields.pop('extras')
        self._data['export_data'].setdefault(name, {})[pk] = fields

    def write_node_repo_folder(self, uuid: str, path: Union[str, Path], overwrite: bool = True):
        self.assert_within_context()
        repo_folder = self._folder.get_subfolder(NODES_EXPORT_SUBFOLDER).get_subfolder(export_shard_uuid(uuid))
        repo_folder.insert_path(src=os.path.abspath(path), dest_name='.', overwrite=overwrite)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for settings and utilities to determine and set the database schema versions."""

import abc
import collections

from aiida.common import exceptions

SCHEMA_VERSION_KEY = 'db|schemaversion'
SCHEMA_VERSION_DESCRIPTION = 'Database schema version'

SCHEMA_GENERATION_KEY = 'schema_generation'  # The key to store the database schema generation in the settings table
SCHEMA_GENERATION_DESCRIPTION = 'Database schema generation'
SCHEMA_GENERATION_VALUE = '1'  # The current schema generation

# Mapping of schema generation onto a tuple of valid schema reset generation and `aiida-core` version number. Given the
# current code schema generation as the key, the first element of the tuple tells what schema generation the database
# should have to be able to reset the schema. If the generation of the database is correct, but the schema version of
# the database does not match the one required for the reset, it means the user first has to downgrade the `aiida-core`
# version and perform the latest migrations. The required version is provided by the tuples second element.
SCHEMA_GENERATION_RESET = {
    '1': ('1', '1.*'),
}

TEMPLATE_INVALID_SCHEMA_GENERATION = """
Database schema generation `{schema_generation_database}` is incompatible with the required schema generation `{schema_generation_code}`.
To migrate the database schema generation to the current one, run the following command:

    verdi -p {profile_name} database migrate
"""

TEMPLATE_INVALID_SCHEMA_VERSION = """
Database schema version `{schema_version_database}` is incompatible with the required schema version `{schema_version_code}`.
To migrate the database schema version to the current one, run the following command:

    verdi -p {profile_name} database migrate
"""

TEMPLATE_MIGRATE_SCHEMA_VERSION_INVALID_VERSION = """
Cannot migrate the database version from `{schema_version_database}` to `{schema_version_code}`.
The database version is ahead of the version of the code and downgrades of the database are not supported.
"""

TEMPLATE_MIGRATE_SCHEMA_GENERATION_INVALID_GENERATION = """
Cannot migrate database schema generation from `{schema_generation_database}` to `{schema_generation_code}`.
This version of `aiida-core` can only migrate databases with schema generation `{schema_generation_reset}`
"""

TEMPLATE_MIGRATE_SCHEMA_GENERATION_INVALID_VERSION = """
Cannot migrate database schema generation from `{schema_generation_database}` to `{schema_generation_code}`.
The current database version is `{schema_version_database}` but `{schema_version_reset}` is required for generation migration.
First install `aiida-core~={aiida_core_version_reset}` and migrate the database to the latest version.
After the database schema is migrated to version `{schema_version_reset}` you can reinstall this version of `aiida-core` and migrate the schema generation.
"""

Setting = collections.namedtuple('Setting', ['key', 'value', 'description', 'time'])


class SettingsManager:
    """Class to get, set and delete settings from the `DbSettings` table."""

    @abc.abstractmethod
    def get(self, key):
        """Return the setting with the given key.

        :param key: the key identifying the setting
        :return: Setting
        :raises: `~aiida.common.exceptions.NotExistent` if the settings does not exist
        """

    @abc.abstractmethod
    def set(self, key, value, description=None):
        """Return the settings with the given key.

        :param key: the key identifying the setting
        :param value: the value for the setting
        :param description: optional setting description
        """

    @abc.abstractmethod
    def delete(self, key):
        """Delete the setting with the given key.

        :param key: the key identifying the setting
        :raises: `~aiida.common.exceptions.NotExistent` if the settings does not exist
        """


class BackendManager:
    """Class to manage the database schema and environment."""

    _settings_manager = None

    @abc.abstractmethod
    def get_settings_manager(self):
        """Return an instance of the `SettingsManager`.

        :return: `SettingsManager`
        """

    def load_backend_environment(self, profile, validate_schema=True, **kwargs):
        """Load the backend environment.

        :param profile: the profile whose backend environment to load
        :param validate_schema: boolean, if True, validate the schema first before loading the environment.
        :param kwargs: keyword arguments that will be passed on to the backend specific scoped session getter function.
        """
        self._load_backend_environment(**kwargs)

        if validate_schema:
            self.validate_schema(profile)

    @abc.abstractmethod
    def _load_backend_environment(self, **kwargs):
        """Load the backend environment.

        :param kwargs: keyword arguments that will be passed on to the backend specific scoped session getter function.
        """

    @abc.abstractmethod
    def reset_backend_environment(self):
        """Reset the backend environment."""

    def migrate(self):
        """Migrate the database to the latest schema generation or version."""
        try:
            # If the settings table does not exist, we are dealing with an empty database. We cannot perform the checks
            # because they rely on the settings table existing, so instead we do not validate but directly call method
            # `_migrate_database_version` which will perform the migration to create the initial schema.
            self.get_settings_manager().validate_table_existence()
        except exceptions.NotExistent:
            self._migrate_database_version()
            return

        if SCHEMA_GENERATION_VALUE != self.get_schema_generation_database():
            self.validate_schema_generation_for_migration()
            self._migrate_database_generation()

        if self.get_schema_version_code() != self.get_schema_version_database():
            self.validate_schema_version_for_migration()
            self._migrate_database_version()

    def _migrate_database_generation(self):
        """Migrate the database schema generation.

        .. warning:: this should NEVER be called directly because there is no validation performed on whether the
            current database schema generation and version can actually be migrated.

        This normally just consists out of setting the schema generation value, but depending on the backend more might
        be needed. In that case, this method should be overridden and call `super` first, followed by the additional
        logic that is required.
        """
        self.set_schema_generation_database(SCHEMA_GENERATION_VALUE)
        self.set_schema_version_database(self.get_schema_version_code())

    def _migrate_database_version(self):
        """Migrate the database to the current schema version.

        .. warning:: this should NEVER be called directly because there is no validation performed on whether the
            current database schema generation and version can actually be migrated.
        """

    @abc.abstractmethod
    def is_database_schema_ahead(self):
        """Determine whether the database schema version is ahead of the code schema version.

        .. warning:: this will not check whether the schema generations are equal

        :return: boolean, True if the database schema version is ahead of the code schema version.
        """

    @abc.abstractmethod
    def get_schema_version_code(self):
        """Return the code schema version."""

    @abc.abstractmethod
    def get_schema_version_reset(self, schema_generation_code):
        """Return schema version the database should have to be able to automatically reset to code schema generation.

        :param schema_generation_code: the schema generation of the code.
        :return: schema version
        """

    @abc.abstractmethod
    def get_schema_version_database(self):
        """Return the database schema version.

        :return: `distutils.version.LooseVersion` with schema version of the database
        """

    @abc.abstractmethod
    def set_schema_version_database(self, version):
        """Set the database schema version.

        :param version: string with schema version to set
        """

    def get_schema_generation_database(self):
        """Return the database schema generation.

        :return: `distutils.version.LooseVersion` with schema generation of the database
        """
        try:
            setting = self.get_settings_manager().get(SCHEMA_GENERATION_KEY)
            return setting.value
        except exceptions.NotExistent:
            return '1'

    def set_schema_generation_database(self, generation):
        """Set the database schema generation.

        :param generation: string with schema generation to set
        """
        self.get_settings_manager().set(SCHEMA_GENERATION_KEY, generation)

    def validate_schema(self, profile):
        """Validate that the current database generation and schema are up-to-date with that of the code.

        :param profile: the profile for which to validate the database schema
        :raises `aiida.common.exceptions.ConfigurationError`: if database schema version or generation is not up-to-date
        """
        self.validate_schema_generation(profile)
        self.validate_schema_version(profile)

    def validate_schema_generation_for_migration(self):
        """Validate whether the current database schema generation can be migrated.

        :raises `aiida.common.exceptions.IncompatibleDatabaseSchema`: if database schema generation cannot be migrated
        """
        schema_generation_code = SCHEMA_GENERATION_VALUE
        schema_generation_database = self.get_schema_generation_database()
        schema_version_database = self.get_schema_version_database()
        schema_version_reset = self.get_schema_version_reset(schema_generation_code)
        schema_generation_reset, aiida_core_version_reset = SCHEMA_GENERATION_RESET[schema_generation_code]

        if schema_generation_database != schema_generation_reset:
            raise exceptions.IncompatibleDatabaseSchema(
                TEMPLATE_MIGRATE_SCHEMA_GENERATION_INVALID_GENERATION.format(
                    schema_generation_database=schema_generation_database,
                    schema_generation_code=schema_generation_code,
                    schema_generation_reset=schema_generation_reset
                )
            )

        if schema_version_database != schema_version_reset:
            raise exceptions.IncompatibleDatabaseSchema(
                TEMPLATE_MIGRATE_SCHEMA_GENERATION_INVALID_VERSION.format(
                    schema_generation_database=schema_generation_database,
                    schema_generation_code=schema_generation_code,
                    schema_version_database=schema_version_database,
                    schema_version_reset=schema_version_reset,
                    aiida_core_version_reset=aiida_core_version_reset
                )
            )

    def validate_schema_version_for_migration(self):
        """Validate whether the current database schema version can be migrated.

        .. warning:: this will not validate that the schema generation is correct.

        :raises `aiida.common.exceptions.IncompatibleDatabaseSchema`: if database schema version cannot be migrated
        """
        schema_version_code = self.get_schema_version_code()
        schema_version_database = self.get_schema_version_database()

        if self.is_database_schema_ahead():
            # Database is newer than the code so a downgrade would be necessary but this is not supported.
            raise exceptions.IncompatibleDatabaseSchema(
                TEMPLATE_MIGRATE_SCHEMA_VERSION_INVALID_VERSION.format(
                    schema_version_database=schema_version_database,
                    schema_version_code=schema_version_code,
                )
            )

    def validate_schema_generation(self, profile):
        """Validate that the current database schema generation is up-to-date with that of the code.

        :raises `aiida.common.exceptions.IncompatibleDatabaseSchema`: if database schema generation is not up-to-date
        """
        schema_generation_code = SCHEMA_GENERATION_VALUE
        schema_generation_database = self.get_schema_generation_database()

        if schema_generation_database != schema_generation_code:
            raise exceptions.IncompatibleDatabaseSchema(
                TEMPLATE_INVALID_SCHEMA_GENERATION.format(
                    schema_generation_database=schema_generation_database,
                    schema_generation_code=schema_generation_code,
                    profile_name=profile.name,
                )
            )

    def validate_schema_version(self, profile):
        """Validate that the current database schema version is up-to-date with that of the code.

        :param profile: the profile for which to validate the database schema
        :raises `aiida.common.exceptions.IncompatibleDatabaseSchema`: if database schema version is not up-to-date
        """
        schema_version_code = self.get_schema_version_code()
        schema_version_database = self.get_schema_version_database()

        if schema_version_database != schema_version_code:
            raise exceptions.IncompatibleDatabaseSchema(
                TEMPLATE_INVALID_SCHEMA_VERSION.format(
                    schema_version_database=schema_version_database,
                    schema_version_code=schema_version_code,
                    profile_name=profile.name
                )
            )



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for all common top level AiiDA entity classes and methods"""
import typing
import abc
import copy

from plumpy.base.utils import super_check, call_with_super_check

from aiida.common import datastructures, exceptions
from aiida.common.lang import classproperty, type_check
from aiida.manage.manager import get_manager

__all__ = ('Entity', 'Collection', 'EntityAttributesMixin', 'EntityExtrasMixin')

EntityType = typing.TypeVar('EntityType')  # pylint: disable=invalid-name

_NO_DEFAULT = tuple()


class Collection(typing.Generic[EntityType]):
    """Container class that represents the collection of objects of a particular type."""

    # A store for any backend specific collections that already exist
    _COLLECTIONS = datastructures.LazyStore()

    @classmethod
    def get_collection(cls, entity_type, backend):
        """
        Get the collection for a given entity type and backend instance

        :param entity_type: the entity type e.g. User, Computer, etc
        :type entity_type: :class:`aiida.orm.Entity`

        :param backend: the backend instance to get the collection for
        :type backend: :class:`aiida.orm.implementation.Backend`

        :return: a new collection with the new backend
        :rtype: :class:`aiida.orm.Collection`
        """
        # Lazily get the collection i.e. create only if we haven't done so yet
        return cls._COLLECTIONS.get((entity_type, backend), lambda: entity_type.Collection(backend, entity_type))

    def __init__(self, backend, entity_class):
        """ Construct a new entity collection.

        :param backend: the backend instance to get the collection for
        :type backend: :class:`aiida.orm.implementation.Backend`

        :param entity_class: the entity type e.g. User, Computer, etc
        :type entity_class: :class:`aiida.orm.Entity`

        """
        assert issubclass(entity_class, Entity), 'Must provide an entity type'
        self._backend = backend or get_manager().get_backend()
        self._entity_type = entity_class

    def __call__(self, backend):
        """ Create a new objects collection using a new backend.

        :param backend: the backend instance to get the collection for
        :type backend: :class:`aiida.orm.implementation.Backend`

        :return: a new collection with the new backend
        :rtype: :class:`aiida.orm.Collection`
        """
        if backend is self._backend:
            # Special case if they actually want the same collection
            return self

        return self.get_collection(self.entity_type, backend)

    @property
    def backend(self):
        """Return the backend.

        :return: the backend instance of this collection
        :rtype: :class:`aiida.orm.implementation.Backend`
        """
        return self._backend

    @property
    def entity_type(self):
        """The entity type.

        :rtype: :class:`aiida.orm.Entity`
        """
        return self._entity_type

    def query(self, filters=None, order_by=None, limit=None, offset=None):
        """
        Get a query builder for the objects of this collection

        :param filters: the keyword value pair filters to match
        :type filters: dict

        :param order_by: a list of (key, direction) pairs specifying the sort order
        :type order_by: list

        :param limit: the maximum number of results to return
        :type limit: int

        :param offset: number of initial results to be skipped
        :type offset: int

        :return: a new query builder instance
        :rtype: :class:`aiida.orm.QueryBuilder`
        """
        from . import querybuilder

        filters = filters or {}
        order_by = {self.entity_type: order_by} if order_by else {}

        query = querybuilder.QueryBuilder(limit=limit, offset=offset)
        query.append(self.entity_type, project='*', filters=filters)
        query.order_by([order_by])
        return query

    def get(self, **filters):
        """
        Get a single collection entry that matches the filter criteria

        :param filters: the filters identifying the object to get
        :type filters: dict

        :return: the entry
        """
        res = self.query(filters=filters)
        return res.one()[0]

    def find(self, filters=None, order_by=None, limit=None):
        """
        Find collection entries matching the filter criteria

        :param filters: the keyword value pair filters to match
        :type filters: dict

        :param order_by: a list of (key, direction) pairs specifying the sort order
        :type order_by: list

        :param limit: the maximum number of results to return
        :type limit: int

        :return: a list of resulting matches
        :rtype: list
        """
        query = self.query(filters=filters, order_by=order_by, limit=limit)
        return query.all(flat=True)

    def all(self):
        """
        Get all entities in this collection

        :return: A list of all entities
        :rtype: list
        """
        return self.query().all(flat=True)  # pylint: disable=no-member

    def count(self, filters=None):
        """Count entities in this collection according to criteria

        :param filters: the keyword value pair filters to match
        :type filters: dict

        :return: The number of entities found using the supplied criteria
        :rtype: int
        """
        return self.query(filters=filters).count()


class Entity:
    """An AiiDA entity"""

    _objects = None

    # Define our collection type
    Collection = Collection

    @classproperty
    def objects(cls, backend=None):  # pylint: disable=no-self-argument
        """
        Get a collection for objects of this type.

        :param backend: the optional backend to use (otherwise use default)
        :type backend: :class:`aiida.orm.implementation.Backend`

        :return: an object that can be used to access entities of this type
        :rtype: :class:`aiida.orm.Collection`
        """
        backend = backend or get_manager().get_backend()
        return cls.Collection.get_collection(cls, backend)

    @classmethod
    def get(cls, **kwargs):
        return cls.objects.get(**kwargs)  # pylint: disable=no-member

    @classmethod
    def from_backend_entity(cls, backend_entity):
        """
        Construct an entity from a backend entity instance

        :param backend_entity: the backend entity

        :return: an AiiDA entity instance
        """
        from .implementation.entities import BackendEntity

        type_check(backend_entity, BackendEntity)
        entity = cls.__new__(cls)
        entity.init_from_backend(backend_entity)
        call_with_super_check(entity.initialize)
        return entity

    def __init__(self, backend_entity):
        """
        :param backend_entity: the backend model supporting this entity
        :type backend_entity: :class:`aiida.orm.implementation.entities.BackendEntity`
        """
        self._backend_entity = backend_entity
        call_with_super_check(self.initialize)

    def init_from_backend(self, backend_entity):
        """
        :param backend_entity: the backend model supporting this entity
        :type backend_entity: :class:`aiida.orm.implementation.entities.BackendEntity`
        """
        self._backend_entity = backend_entity

    @super_check
    def initialize(self):
        """Initialize instance attributes.

        This will be called after the constructor is called or an entity is created from an existing backend entity.
        """

    @property
    def id(self):  # pylint: disable=invalid-name
        """Return the id for this entity.

        This identifier is guaranteed to be unique amongst entities of the same type for a single backend instance.

        :return: the entity's id
        """
        return self._backend_entity.id

    @property
    def pk(self):
        """Return the primary key for this entity.

        This identifier is guaranteed to be unique amongst entities of the same type for a single backend instance.

        :return: the entity's principal key
        """
        return self.id

    @property
    def uuid(self):
        """Return the UUID for this entity.

        This identifier is unique across all entities types and backend instances.

        :return: the entity uuid
        :rtype: :class:`uuid.UUID`
        """
        return self._backend_entity.uuid

    def store(self):
        """Store the entity."""
        self._backend_entity.store()
        return self

    @property
    def is_stored(self):
        """Return whether the entity is stored.

        :return: boolean, True if stored, False otherwise
        :rtype: bool
        """
        return self._backend_entity.is_stored

    @property
    def backend(self):
        """
        Get the backend for this entity
        :return: the backend instance
        """
        return self._backend_entity.backend

    @property
    def backend_entity(self):
        """
        Get the implementing class for this object

        :return: the class model
        """
        return self._backend_entity


class EntityAttributesMixin(abc.ABC):
    """Mixin class that adds all methods for the attributes column to an entity."""

    @property
    def attributes(self):
        """Return the complete attributes dictionary.

        .. warning:: While the entity is unstored, this will return references of the attributes on the database model,
            meaning that changes on the returned values (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            attributes will be a deep copy and mutations of the database attributes will have to go through the
            appropriate set methods. Therefore, once stored, retrieving a deep copy can be a heavy operation. If you
            only need the keys or some values, use the iterators `attributes_keys` and `attributes_items`, or the
            getters `get_attribute` and `get_attribute_many` instead.

        :return: the attributes as a dictionary
        """
        attributes = self.backend_entity.attributes

        if self.is_stored:
            attributes = copy.deepcopy(attributes)

        return attributes

    def get_attribute(self, key, default=_NO_DEFAULT):
        """Return the value of an attribute.

        .. warning:: While the entity is unstored, this will return a reference of the attribute on the database model,
            meaning that changes on the returned value (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            attribute will be a deep copy and mutations of the database attributes will have to go through the
            appropriate set methods.

        :param key: name of the attribute
        :param default: return this value instead of raising if the attribute does not exist
        :return: the value of the attribute
        :raises AttributeError: if the attribute does not exist and no default is specified
        """
        try:
            attribute = self.backend_entity.get_attribute(key)
        except AttributeError:
            if default is _NO_DEFAULT:
                raise
            attribute = default

        if self.is_stored:
            attribute = copy.deepcopy(attribute)

        return attribute

    def get_attribute_many(self, keys):
        """Return the values of multiple attributes.

        .. warning:: While the entity is unstored, this will return references of the attributes on the database model,
            meaning that changes on the returned values (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            attributes will be a deep copy and mutations of the database attributes will have to go through the
            appropriate set methods. Therefore, once stored, retrieving a deep copy can be a heavy operation. If you
            only need the keys or some values, use the iterators `attributes_keys` and `attributes_items`, or the
            getters `get_attribute` and `get_attribute_many` instead.

        :param keys: a list of attribute names
        :return: a list of attribute values
        :raises AttributeError: if at least one attribute does not exist
        """
        attributes = self.backend_entity.get_attribute_many(keys)

        if self.is_stored:
            attributes = copy.deepcopy(attributes)

        return attributes

    def set_attribute(self, key, value):
        """Set an attribute to the given value.

        :param key: name of the attribute
        :param value: value of the attribute
        :raise aiida.common.ValidationError: if the key is invalid, i.e. contains periods
        :raise aiida.common.ModificationNotAllowed: if the entity is stored
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the attributes of a stored entity are immutable')

        self.backend_entity.set_attribute(key, value)

    def set_attribute_many(self, attributes):
        """Set multiple attributes.

        .. note:: This will override any existing attributes that are present in the new dictionary.

        :param attributes: a dictionary with the attributes to set
        :raise aiida.common.ValidationError: if any of the keys are invalid, i.e. contain periods
        :raise aiida.common.ModificationNotAllowed: if the entity is stored
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the attributes of a stored entity are immutable')

        self.backend_entity.set_attribute_many(attributes)

    def reset_attributes(self, attributes):
        """Reset the attributes.

        .. note:: This will completely clear any existing attributes and replace them with the new dictionary.

        :param attributes: a dictionary with the attributes to set
        :raise aiida.common.ValidationError: if any of the keys are invalid, i.e. contain periods
        :raise aiida.common.ModificationNotAllowed: if the entity is stored
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the attributes of a stored entity are immutable')

        self.backend_entity.reset_attributes(attributes)

    def delete_attribute(self, key):
        """Delete an attribute.

        :param key: name of the attribute
        :raises AttributeError: if the attribute does not exist
        :raise aiida.common.ModificationNotAllowed: if the entity is stored
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the attributes of a stored entity are immutable')

        self.backend_entity.delete_attribute(key)

    def delete_attribute_many(self, keys):
        """Delete multiple attributes.

        :param keys: names of the attributes to delete
        :raises AttributeError: if at least one of the attribute does not exist
        :raise aiida.common.ModificationNotAllowed: if the entity is stored
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the attributes of a stored entity are immutable')

        self.backend_entity.delete_attribute_many(keys)

    def clear_attributes(self):
        """Delete all attributes."""
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the attributes of a stored entity are immutable')

        self.backend_entity.clear_attributes()

    def attributes_items(self):
        """Return an iterator over the attributes.

        :return: an iterator with attribute key value pairs
        """
        return self.backend_entity.attributes_items()

    def attributes_keys(self):
        """Return an iterator over the attribute keys.

        :return: an iterator with attribute keys
        """
        return self.backend_entity.attributes_keys()


class EntityExtrasMixin(abc.ABC):
    """Mixin class that adds all methods for the extras column to an entity."""

    @property
    def extras(self):
        """Return the complete extras dictionary.

        .. warning:: While the entity is unstored, this will return references of the extras on the database model,
            meaning that changes on the returned values (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            extras will be a deep copy and mutations of the database extras will have to go through the appropriate set
            methods. Therefore, once stored, retrieving a deep copy can be a heavy operation. If you only need the keys
            or some values, use the iterators `extras_keys` and `extras_items`, or the getters `get_extra` and
            `get_extra_many` instead.

        :return: the extras as a dictionary
        """
        extras = self.backend_entity.extras

        if self.is_stored:
            extras = copy.deepcopy(extras)

        return extras

    def get_extra(self, key, default=_NO_DEFAULT):
        """Return the value of an extra.

        .. warning:: While the entity is unstored, this will return a reference of the extra on the database model,
            meaning that changes on the returned value (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            extra will be a deep copy and mutations of the database extras will have to go through the appropriate set
            methods.

        :param key: name of the extra
        :param default: return this value instead of raising if the attribute does not exist
        :return: the value of the extra
        :raises AttributeError: if the extra does not exist and no default is specified
        """
        try:
            extra = self.backend_entity.get_extra(key)
        except AttributeError:
            if default is _NO_DEFAULT:
                raise
            extra = default

        if self.is_stored:
            extra = copy.deepcopy(extra)

        return extra

    def get_extra_many(self, keys):
        """Return the values of multiple extras.

        .. warning:: While the entity is unstored, this will return references of the extras on the database model,
            meaning that changes on the returned values (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            extras will be a deep copy and mutations of the database extras will have to go through the appropriate set
            methods. Therefore, once stored, retrieving a deep copy can be a heavy operation. If you only need the keys
            or some values, use the iterators `extras_keys` and `extras_items`, or the getters `get_extra` and
            `get_extra_many` instead.

        :param keys: a list of extra names
        :return: a list of extra values
        :raises AttributeError: if at least one extra does not exist
        """
        extras = self.backend_entity.get_extra_many(keys)

        if self.is_stored:
            extras = copy.deepcopy(extras)

        return extras

    def set_extra(self, key, value):
        """Set an extra to the given value.

        :param key: name of the extra
        :param value: value of the extra
        :raise aiida.common.ValidationError: if the key is invalid, i.e. contains periods
        """
        self.backend_entity.set_extra(key, value)

    def set_extra_many(self, extras):
        """Set multiple extras.

        .. note:: This will override any existing extras that are present in the new dictionary.

        :param extras: a dictionary with the extras to set
        :raise aiida.common.ValidationError: if any of the keys are invalid, i.e. contain periods
        """
        self.backend_entity.set_extra_many(extras)

    def reset_extras(self, extras):
        """Reset the extras.

        .. note:: This will completely clear any existing extras and replace them with the new dictionary.

        :param extras: a dictionary with the extras to set
        :raise aiida.common.ValidationError: if any of the keys are invalid, i.e. contain periods
        """
        self.backend_entity.reset_extras(extras)

    def delete_extra(self, key):
        """Delete an extra.

        :param key: name of the extra
        :raises AttributeError: if the extra does not exist
        """
        self.backend_entity.delete_extra(key)

    def delete_extra_many(self, keys):
        """Delete multiple extras.

        :param keys: names of the extras to delete
        :raises AttributeError: if at least one of the extra does not exist
        """
        self.backend_entity.delete_extra_many(keys)

    def clear_extras(self):
        """Delete all extras."""
        self.backend_entity.clear_extras()

    def extras_items(self):
        """Return an iterator over the extras.

        :return: an iterator with extra key value pairs
        """
        return self.backend_entity.extras_items()

    def extras_keys(self):
        """Return an iterator over the extra keys.

        :return: an iterator with extra keys
        """
        return self.backend_entity.extras_keys()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi data dict` command."""

from aiida.cmdline.commands.cmd_data import verdi_data
from aiida.cmdline.params import arguments, options, types


@verdi_data.group('dict')
def dictionary():
    """Manipulate Dict objects (python dictionaries)."""


@dictionary.command('show')
@arguments.DATA(type=types.DataParamType(sub_classes=('aiida.data:dict',)))
@options.DICT_FORMAT()
def dictionary_show(data, fmt):
    """Show contents of Dict nodes."""
    from aiida.cmdline.utils.echo import echo_dictionary

    for node in data:
        echo_dictionary(node.get_dict(), fmt)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Abstract BackendNode and BackendNodeCollection implementation."""

import abc

from .entities import BackendEntity, BackendCollection, BackendEntityAttributesMixin, BackendEntityExtrasMixin

__all__ = ('BackendNode', 'BackendNodeCollection')


class BackendNode(BackendEntity, BackendEntityExtrasMixin, BackendEntityAttributesMixin, metaclass=abc.ABCMeta):
    """Wrapper around a `DbNode` instance to set and retrieve data independent of the database implementation."""

    # pylint: disable=too-many-public-methods

    @abc.abstractmethod
    def clone(self):
        """Return an unstored clone of ourselves.

        :return: an unstored `BackendNode` with the exact same attributes and extras as self
        """

    @property
    def uuid(self):
        """Return the node UUID.

        :return: the string representation of the UUID
        :rtype: str or None
        """
        if self._dbmodel.uuid:
            return str(self._dbmodel.uuid)

        return None

    @property
    def node_type(self):
        """Return the node type.

        :return: the node type
        """
        return self._dbmodel.node_type

    @property
    def process_type(self):
        """Return the node process type.

        :return: the process type
        """
        return self._dbmodel.process_type

    @process_type.setter
    def process_type(self, value):
        """Set the process type.

        :param value: the new value to set
        """
        self._dbmodel.process_type = value

    @property
    def label(self):
        """Return the node label.

        :return: the label
        """
        return self._dbmodel.label

    @label.setter
    def label(self, value):
        """Set the label.

        :param value: the new value to set
        """
        self._dbmodel.label = value

    @property
    def description(self):
        """Return the node description.

        :return: the description
        """
        return self._dbmodel.description

    @description.setter
    def description(self, value):
        """Set the description.

        :param value: the new value to set
        """
        self._dbmodel.description = value

    @abc.abstractproperty
    def computer(self):
        """Return the computer of this node.

        :return: the computer or None
        :rtype: `BackendComputer` or None
        """

    @computer.setter
    @abc.abstractmethod
    def computer(self, computer):
        """Set the computer of this node.

        :param computer: a `BackendComputer`
        """

    @abc.abstractproperty
    def user(self):
        """Return the user of this node.

        :return: the user
        :rtype: `BackendUser`
        """

    @user.setter
    @abc.abstractmethod
    def user(self, user):
        """Set the user of this node.

        :param user: a `BackendUser`
        """

    @property
    def ctime(self):
        """Return the node ctime.

        :return: the ctime
        """
        return self._dbmodel.ctime

    @property
    def mtime(self):
        """Return the node mtime.

        :return: the mtime
        """
        return self._dbmodel.mtime

    @abc.abstractmethod
    def add_incoming(self, source, link_type, link_label):
        """Add a link of the given type from a given node to ourself.

        :param source: the node from which the link is coming
        :param link_type: the link type
        :param link_label: the link label
        :return: True if the proposed link is allowed, False otherwise
        :raise TypeError: if `source` is not a Node instance or `link_type` is not a `LinkType` enum
        :raise ValueError: if the proposed link is invalid
        """

    @abc.abstractmethod
    def store(self, links=None, with_transaction=True, clean=True):  # pylint: disable=arguments-differ
        """Store the node in the database.

        :param links: optional links to add before storing
        :param with_transaction: if False, do not use a transaction because the caller will already have opened one.
        :param clean: boolean, if True, will clean the attributes and extras before attempting to store
        """


class BackendNodeCollection(BackendCollection[BackendNode]):
    """The collection of `BackendNode` entries."""

    ENTITY_CLASS = BackendNode

    @abc.abstractmethod
    def get(self, pk):
        """Return a Node entry from the collection with the given id

        :param pk: id of the node
        """

    @abc.abstractmethod
    def delete(self, pk):
        """Remove a Node entry from the collection with the given id

        :param pk: id of the node to delete
        """



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""SqlAlchemy implementation of the `BackendNode` and `BackendNodeCollection` classes."""

# pylint: disable=no-name-in-module,import-error
from datetime import datetime
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.exc import SQLAlchemyError

from aiida.backends.sqlalchemy import get_scoped_session
from aiida.backends.sqlalchemy.models import node as models
from aiida.common import exceptions
from aiida.common.lang import type_check
from aiida.orm.implementation.utils import clean_value

from .. import BackendNode, BackendNodeCollection
from . import entities
from . import utils as sqla_utils
from .computers import SqlaComputer
from .users import SqlaUser


class SqlaNode(entities.SqlaModelEntity[models.DbNode], BackendNode):
    """SQLA Node backend entity"""

    # pylint: disable=too-many-public-methods

    MODEL_CLASS = models.DbNode

    def __init__(
        self,
        backend,
        node_type,
        user,
        computer=None,
        process_type=None,
        label='',
        description='',
        ctime=None,
        mtime=None
    ):
        """Construct a new `BackendNode` instance wrapping a new `DbNode` instance.

        :param backend: the backend
        :param node_type: the node type string
        :param user: associated `BackendUser`
        :param computer: associated `BackendComputer`
        :param label: string label
        :param description: string description
        :param ctime: The creation time as datetime object
        :param mtime: The modification time as datetime object
        """
        # pylint: disable=too-many-arguments
        super().__init__(backend)

        arguments = {
            'node_type': node_type,
            'process_type': process_type,
            'user': user.dbmodel,
            'label': label,
            'description': description,
        }

        type_check(user, SqlaUser)

        if computer:
            type_check(computer, SqlaComputer, f'computer is of type {type(computer)}')
            arguments['dbcomputer'] = computer.dbmodel

        if ctime:
            type_check(ctime, datetime, f'the given ctime is of type {type(ctime)}')
            arguments['ctime'] = ctime

        if mtime:
            type_check(mtime, datetime, f'the given mtime is of type {type(mtime)}')
            arguments['mtime'] = mtime

        self._dbmodel = sqla_utils.ModelWrapper(models.DbNode(**arguments))

    def clone(self):
        """Return an unstored clone of ourselves.

        :return: an unstored `BackendNode` with the exact same attributes and extras as self
        """
        arguments = {
            'node_type': self._dbmodel.node_type,
            'process_type': self._dbmodel.process_type,
            'user': self._dbmodel.user,
            'dbcomputer': self._dbmodel.dbcomputer,
            'label': self._dbmodel.label,
            'description': self._dbmodel.description,
            'attributes': self._dbmodel.attributes,
            'extras': self._dbmodel.extras,
        }

        clone = self.__class__.__new__(self.__class__)  # pylint: disable=no-value-for-parameter
        clone.__init__(self.backend, self.node_type, self.user)
        clone._dbmodel = sqla_utils.ModelWrapper(models.DbNode(**arguments))  # pylint: disable=protected-access
        return clone

    @property
    def computer(self):
        """Return the computer of this node.

        :return: the computer or None
        :rtype: `BackendComputer` or None
        """
        try:
            return self.backend.computers.from_dbmodel(self._dbmodel.dbcomputer)
        except TypeError:
            return None

    @computer.setter
    def computer(self, computer):
        """Set the computer of this node.

        :param computer: a `BackendComputer`
        """
        type_check(computer, SqlaComputer, allow_none=True)

        if computer is not None:
            computer = computer.dbmodel

        self._dbmodel.dbcomputer = computer

    @property
    def user(self):
        """Return the user of this node.

        :return: the user
        :rtype: `BackendUser`
        """
        return self.backend.users.from_dbmodel(self._dbmodel.user)

    @user.setter
    def user(self, user):
        """Set the user of this node.

        :param user: a `BackendUser`
        """
        type_check(user, SqlaUser)
        self._dbmodel.user = user.dbmodel

    def add_incoming(self, source, link_type, link_label):
        """Add a link of the given type from a given node to ourself.

        :param source: the node from which the link is coming
        :param link_type: the link type
        :param link_label: the link label
        :return: True if the proposed link is allowed, False otherwise
        :raise aiida.common.ModificationNotAllowed: if either source or target node is not stored
        """
        session = get_scoped_session()

        type_check(source, SqlaNode)

        if not self.is_stored:
            raise exceptions.ModificationNotAllowed('node has to be stored when adding an incoming link')

        if not source.is_stored:
            raise exceptions.ModificationNotAllowed('source node has to be stored when adding a link from it')

        self._add_link(source, link_type, link_label)
        session.commit()

    def _add_link(self, source, link_type, link_label):
        """Add a link of the given type from a given node to ourself.

        :param source: the node from which the link is coming
        :param link_type: the link type
        :param link_label: the link label
        """
        from aiida.backends.sqlalchemy.models.node import DbLink

        session = get_scoped_session()

        try:
            with session.begin_nested():
                link = DbLink(input_id=source.id, output_id=self.id, label=link_label, type=link_type.value)
                session.add(link)
        except SQLAlchemyError as exception:
            raise exceptions.UniquenessError(f'failed to create the link: {exception}') from exception

    def clean_values(self):
        self._dbmodel.attributes = clean_value(self._dbmodel.attributes)
        self._dbmodel.extras = clean_value(self._dbmodel.extras)

    def store(self, links=None, with_transaction=True, clean=True):  # pylint: disable=arguments-differ
        """Store the node in the database.

        :param links: optional links to add before storing
        :param with_transaction: if False, do not use a transaction because the caller will already have opened one.
        :param clean: boolean, if True, will clean the attributes and extras before attempting to store
        """
        session = get_scoped_session()

        if clean:
            self.clean_values()

        session.add(self._dbmodel)

        if links:
            for link_triple in links:
                self._add_link(*link_triple)

        if with_transaction:
            try:
                session.commit()
            except SQLAlchemyError:
                session.rollback()
                raise

        return self


class SqlaNodeCollection(BackendNodeCollection):
    """The collection of Node entries."""

    ENTITY_CLASS = SqlaNode

    def get(self, pk):
        """Return a Node entry from the collection with the given id

        :param pk: id of the node
        """
        session = get_scoped_session()

        try:
            return self.ENTITY_CLASS.from_dbmodel(session.query(models.DbNode).filter_by(id=pk).one(), self.backend)
        except NoResultFound:
            raise exceptions.NotExistent(f"Node with pk '{pk}' not found") from NoResultFound

    def delete(self, pk):
        """Remove a Node entry from the collection with the given id

        :param pk: id of the node to delete
        """
        session = get_scoped_session()

        try:
            session.query(models.DbNode).filter_by(id=pk).one().delete()
            session.commit()
        except NoResultFound:
            raise exceptions.NotExistent(f"Node with pk '{pk}' not found") from NoResultFound



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=too-many-branches
"""Utility functions and classes to interact with AiiDA export archives."""
import os
import sys
import tarfile
import zipfile
import warnings

from aiida.common.progress_reporter import get_progress_reporter, set_progress_bar_tqdm, set_progress_reporter
from aiida.common.warnings import AiidaDeprecationWarning
from aiida.tools.importexport.common.config import NODES_EXPORT_SUBFOLDER
from aiida.tools.importexport.common.exceptions import CorruptArchive

__all__ = ('extract_zip', 'extract_tar', 'extract_tree')


def _update_description(path, progress_bar, refresh: bool = False):
    """Update description for a progress bar given path

    :param path: path of file or directory
    :type path: str
    :param refresh: Whether or not to refresh the progress bar with the new description. Default: False.
    :type refresh: bool
    """
    (path, description) = os.path.split(path)
    while description == '':
        (path, description) = os.path.split(path)
    description = f'EXTRACTING: {description}'

    progress_bar.set_description_str(description, refresh=refresh)


def extract_zip(infile, folder, nodes_export_subfolder=None, check_files=('data.json', 'metadata.json'), **kwargs):
    """Extract the nodes to be imported from a zip file.

    :param infile: file path
    :type infile: str

    :param folder: a temporary folder used to extract the file tree
    :type folder: :py:class:`~aiida.common.folders.SandboxFolder`

    :param nodes_export_subfolder: name of the subfolder for AiiDA nodes
    :type nodes_export_subfolder: str

    :param check_files: list of files to check are present

    :param silent: suppress progress bar
    :type silent: bool

    :raises TypeError: if parameter types are not respected
    :raises `~aiida.tools.importexport.common.exceptions.CorruptArchive`: if the archive misses files or files have
        incorrect formats
    """
    warnings.warn(
        'extract_zip function is deprecated and will be removed in AiiDA v2.0.0, '
        'use extract_tree in the archive-path package instead', AiidaDeprecationWarning
    )  # pylint: disable=no-member

    if nodes_export_subfolder:
        if not isinstance(nodes_export_subfolder, str):
            raise TypeError('nodes_export_subfolder must be a string')
    else:
        nodes_export_subfolder = NODES_EXPORT_SUBFOLDER

    if not kwargs.get('silent', False):
        set_progress_bar_tqdm(unit='files')
    else:
        set_progress_reporter(None)

    data_files = set()

    try:
        with zipfile.ZipFile(infile, 'r', allowZip64=True) as handle:

            members = handle.namelist()

            if not members:
                raise CorruptArchive('no files detected in archive')

            with get_progress_reporter()(total=len(members)) as progress:

                for membername in members:

                    progress.update()

                    # Check that we are only exporting nodes within the subfolder!
                    # better check such that there are no .. in the
                    # path; use probably the folder limit checks
                    if membername in check_files:
                        data_files.add(membername)
                    elif not membername.startswith(nodes_export_subfolder + os.sep):
                        continue

                    _update_description(membername, progress)

                    handle.extract(path=folder.abspath, member=membername)

    except zipfile.BadZipfile:
        raise ValueError('The input file format for import is not valid (not a zip file)')

    for name in check_files:
        if name not in data_files:
            raise CorruptArchive('Archive missing required file:f {name}')


def extract_tar(infile, folder, nodes_export_subfolder=None, check_files=('data.json', 'metadata.json'), **kwargs):
    """
    Extract the nodes to be imported from a (possibly zipped) tar file.

    :param infile: file path
    :type infile: str

    :param folder: a temporary fodler used to extract the file tree
    :type folder: :py:class:`~aiida.common.folders.SandboxFolder`

    :param nodes_export_subfolder: name of the subfolder for AiiDA nodes
    :type nodes_export_subfolder: str

    :param check_files: list of files to check are present

    :param silent: suppress progress bar
    :type silent: bool

    :raises TypeError: if parameter types are not respected
    :raises `~aiida.tools.importexport.common.exceptions.CorruptArchive`: if the archive misses files or files have
        incorrect formats
    """
    warnings.warn(
        'extract_tar function is deprecated and will be removed in AiiDA v2.0.0, '
        'use extract_tree in the archive-path package instead', AiidaDeprecationWarning
    )  # pylint: disable=no-member

    if nodes_export_subfolder:
        if not isinstance(nodes_export_subfolder, str):
            raise TypeError('nodes_export_subfolder must be a string')
    else:
        nodes_export_subfolder = NODES_EXPORT_SUBFOLDER

    if not kwargs.get('silent', False):
        set_progress_bar_tqdm(unit='files')
    else:
        set_progress_reporter(None)

    data_files = set()

    try:
        with tarfile.open(infile, 'r:*', format=tarfile.PAX_FORMAT) as handle:

            members = handle.getmembers()

            if len(members) == 1 and members[0].size == 0:
                raise CorruptArchive('no files detected in archive')

            with get_progress_reporter()(total=len(members)) as progress:

                for member in members:

                    progress.update()

                    if member.isdev():
                        # safety: skip if character device, block device or FIFO
                        print(f'WARNING, device found inside the import file: {member.name}', file=sys.stderr)
                        continue
                    if member.issym() or member.islnk():
                        # safety: in export, I set dereference=True therefore
                        # there should be no symbolic or hard links.
                        print(f'WARNING, symlink found inside the import file: {member.name}', file=sys.stderr)
                        continue
                    # Check that we are only exporting nodes within the subfolder!
                    # better check such that there are no .. in the
                    # path; use probably the folder limit checks
                    if member.name in check_files:
                        data_files.add(member.name)
                    elif not member.name.startswith(nodes_export_subfolder + os.sep):
                        continue

                    _update_description(member.name, progress)

                    handle.extract(path=folder.abspath, member=member)
    except tarfile.ReadError:
        raise ValueError('The input file format for import is not valid (not a tar file)')

    for name in check_files:
        if name not in data_files:
            raise CorruptArchive('Archive missing required file:f {name}')


def extract_tree(infile, folder):
    """Prepare to import nodes from plain file system tree by copying in the given sandbox folder.

    .. note:: the contents of the unpacked archive directory are copied into the sandbox folder, because the files will
        anyway haven to be copied to the repository at some point. By copying the contents of the source directory now
        and continuing operation only on the sandbox folder, we do not risk to modify the source files accidentally.
        During import then, the node files from the sandbox can be moved to the repository, so they won't have to be
        copied again in any case.

    :param infile: absolute filepath point to the unpacked archive directory
    :type infile: str

    :param folder: a temporary folder to which the archive contents are copied
    :type folder: :py:class:`~aiida.common.folders.SandboxFolder`
    """
    folder.replace_with_folder(infile, move=False, overwrite=True)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Translator for bands data
"""

from aiida.restapi.translator.nodes.data import DataTranslator


class BandsDataTranslator(DataTranslator):
    """
    Translator relative to resource 'bands' and aiida class BandsData
    """

    # A label associated to the present class (coincides with the resource name)
    __label__ = 'bands'
    # The AiiDA class one-to-one associated to the present class
    from aiida.orm import BandsData
    _aiida_class = BandsData
    # The string name of the AiiDA class
    _aiida_type = 'data.array.bands.BandsData'

    _result_type = __label__

    @staticmethod
    def get_derived_properties(node):
        """
        Returns: data in a format required by dr.js to visualize a 2D plot
        with multiple data series.

        Strategy: I take advantage of the export functionality of BandsData
        objects. The raw export has to be filtered for string escape characters.
        this is done by decoding the string returned by node._exportcontent.

        TODO: modify the function exportstring (or add another function in
        BandsData) so that it returns a python object rather than a string.
        """
        response = {}

        from aiida.common import json
        json_string = node._exportcontent('json', comments=False)  # pylint: disable=protected-access
        json_content = json.loads(json_string[0])
        response['bands'] = json_content

        return response



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name
"""Database migration."""
from django.db import migrations
from aiida.backends.djsite.db.migrations import upgrade_schema_version

REVISION = '1.0.8'
DOWN_REVISION = '1.0.7'


class Migration(migrations.Migration):
    """Database migration."""

    dependencies = [
        ('db', '0007_update_linktypes'),
    ]

    operations = [
        # The 'hidden' property of AbstractCode has been changed from an attribute to an extra
        # Therefore we find all nodes of type Code and if they have an attribute with the key 'hidden'
        # we move that value to the extra table
        #
        # First we copy the 'hidden' attributes from code.Code. nodes to the db_extra table
        migrations.RunSQL(
            """
            INSERT INTO db_dbextra (key, datatype, tval, fval, ival, bval, dval, dbnode_id) (
                SELECT db_dbattribute.key, db_dbattribute.datatype, db_dbattribute.tval, db_dbattribute.fval,
                db_dbattribute.ival, db_dbattribute.bval, db_dbattribute.dval, db_dbattribute.dbnode_id
                FROM db_dbattribute JOIN db_dbnode ON db_dbnode.id = db_dbattribute.dbnode_id
                WHERE db_dbattribute.key = 'hidden'
                    AND db_dbnode.type = 'code.Code.'
            );
        """
        ),
        # Secondly, we delete the original entries from the DbAttribute table
        migrations.RunSQL(
            """
            DELETE FROM db_dbattribute
            WHERE id in (
                SELECT db_dbattribute.id
                FROM db_dbattribute
                JOIN db_dbnode ON db_dbnode.id = db_dbattribute.dbnode_id
                WHERE db_dbattribute.key = 'hidden' AND db_dbnode.type = 'code.Code.'
            );
        """
        ),
        upgrade_schema_version(REVISION, DOWN_REVISION)
    ]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Basic test classes."""
import os
import unittest
import traceback

from tornado import ioloop

from aiida.common.exceptions import ConfigurationError, TestsNotAllowedError, InternalError
from aiida.common.lang import classproperty
from aiida.manage import configuration
from aiida.manage.manager import get_manager, reset_manager

TEST_KEYWORD = 'test_'


def check_if_tests_can_run():
    """Verify that the currently loaded profile is a test profile, otherwise raise `TestsNotAllowedError`."""
    profile = configuration.PROFILE
    if not profile.is_test_profile:
        raise TestsNotAllowedError(f'currently loaded profile {profile.name} is not a valid test profile')


class AiidaTestCase(unittest.TestCase):
    """This is the base class for AiiDA tests, independent of the backend.

    Internally it loads the AiidaTestImplementation subclass according to the current backend."""
    _class_was_setup = False
    __backend_instance = None
    backend = None  # type: aiida.orm.implementation.Backend

    @classmethod
    def get_backend_class(cls):
        """Get backend class."""
        from aiida.backends.testimplbase import AiidaTestImplementation
        from aiida.backends import BACKEND_SQLA, BACKEND_DJANGO
        from aiida.manage.configuration import PROFILE

        # Freeze the __impl_class after the first run
        if not hasattr(cls, '__impl_class'):
            if PROFILE.database_backend == BACKEND_SQLA:
                from aiida.backends.sqlalchemy.testbase import SqlAlchemyTests
                cls.__impl_class = SqlAlchemyTests
            elif PROFILE.database_backend == BACKEND_DJANGO:
                from aiida.backends.djsite.db.testbase import DjangoTests
                cls.__impl_class = DjangoTests
            else:
                raise ConfigurationError('Unknown backend type')

            # Check that it is of the right class
            if not issubclass(cls.__impl_class, AiidaTestImplementation):
                raise InternalError(
                    'The AiiDA test implementation is not of type '
                    '{}, that is not a subclass of AiidaTestImplementation'.format(cls.__impl_class.__name__)
                )

        return cls.__impl_class

    @classmethod
    def setUpClass(cls, *args, **kwargs):  # pylint: disable=arguments-differ
        # Note: this will raise an exception, that will be seen as a test
        # failure. To be safe, you should do the same check also in the tearDownClass
        # to avoid that it is run
        check_if_tests_can_run()

        # Force the loading of the backend which will load the required database environment
        get_manager().get_backend()

        cls.__backend_instance = cls.get_backend_class()()
        cls.__backend_instance.setUpClass_method(*args, **kwargs)
        cls.backend = cls.__backend_instance.backend

        cls._class_was_setup = True

        cls.clean_db()
        cls.insert_data()

    def setUp(self):
        # Install a new IOLoop so that any messing up of the state of the loop is not propagated
        # to subsequent tests.
        # This call should come before the backend instance setup call just in case it uses the loop
        ioloop.IOLoop().make_current()

    def tearDown(self):
        # Clean up the loop we created in set up.
        # Call this after the instance tear down just in case it uses the loop
        reset_manager()
        loop = ioloop.IOLoop.current()
        if not loop._closing:  # pylint: disable=protected-access,no-member
            loop.close()

    def reset_database(self):
        """Reset the database to the default state deleting any content currently stored"""
        from aiida.orm import autogroup

        self.clean_db()
        if autogroup.CURRENT_AUTOGROUP is not None:
            autogroup.CURRENT_AUTOGROUP.clear_group_cache()
        self.insert_data()

    @classmethod
    def insert_data(cls):
        """
        This method setups the database (by creating a default user) and
        inserts default data into the database (which is for the moment a
        default computer).
        """
        from aiida.orm import User

        cls.create_user()
        User.objects.reset()
        cls.create_computer()

    @classmethod
    def create_user(cls):
        cls.__backend_instance.create_user()

    @classmethod
    def create_computer(cls):
        cls.__backend_instance.create_computer()

    @classmethod
    def clean_db(cls):
        """Clean up database and reset caches.

        Resets AiiDA manager cache, which could otherwise be left in an inconsistent state when cleaning the database.
        """
        from aiida.common.exceptions import InvalidOperation

        # Note: this will raise an exception, that will be seen as a test
        # failure. To be safe, you should do the same check also in the tearDownClass
        # to avoid that it is run
        check_if_tests_can_run()

        if not cls._class_was_setup:
            raise InvalidOperation('You cannot call clean_db before running the setUpClass')

        cls.__backend_instance.clean_db()

        reset_manager()

    @classmethod
    def clean_repository(cls):
        """
        Cleans up file repository.
        """
        from aiida.manage.configuration import get_profile
        from aiida.common.exceptions import InvalidOperation
        import shutil

        dirpath_repository = get_profile().repository_path

        base_repo_path = os.path.basename(os.path.normpath(dirpath_repository))
        if TEST_KEYWORD not in base_repo_path:
            raise InvalidOperation(
                'Warning: The repository folder {} does not '
                'seem to belong to a test profile and will therefore not be deleted.\n'
                'Full repository path: '
                '{}'.format(base_repo_path, dirpath_repository)
            )

        # Clean the test repository
        shutil.rmtree(dirpath_repository, ignore_errors=True)
        os.makedirs(dirpath_repository)

    @classproperty
    def computer(cls):  # pylint: disable=no-self-argument
        """Get the default computer for this test

        :return: the test computer
        :rtype: :class:`aiida.orm.Computer`"""
        return cls.__backend_instance.get_computer()

    @classproperty
    def user_email(cls):  # pylint: disable=no-self-argument
        return cls.__backend_instance.get_user_email()

    @classmethod
    def tearDownClass(cls, *args, **kwargs):  # pylint: disable=arguments-differ
        # Double check for double security to avoid to run the tearDown
        # if this is not a test profile
        from aiida.orm import autogroup

        check_if_tests_can_run()
        if autogroup.CURRENT_AUTOGROUP is not None:
            autogroup.CURRENT_AUTOGROUP.clear_group_cache()
        cls.clean_db()
        cls.clean_repository()
        cls.__backend_instance.tearDownClass_method(*args, **kwargs)

    def assertClickSuccess(self, cli_result):  # pylint: disable=invalid-name
        self.assertEqual(cli_result.exit_code, 0, cli_result.output)
        self.assertClickResultNoException(cli_result)

    def assertClickResultNoException(self, cli_result):  # pylint: disable=invalid-name
        self.assertIsNone(cli_result.exception, ''.join(traceback.format_exception(*cli_result.exc_info)))


class AiidaPostgresTestCase(AiidaTestCase):
    """Setup postgres tests."""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        """Setup the PGTest postgres test cluster."""
        from pgtest.pgtest import PGTest
        cls.pg_test = PGTest()
        super().setUpClass(*args, **kwargs)

    @classmethod
    def tearDownClass(cls, *args, **kwargs):
        """Close the PGTest postgres test cluster."""
        super().tearDownClass(*args, **kwargs)
        cls.pg_test.close()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Defines an rst directive to auto-document AiiDA calculation job.
"""
import inspect

from aiida.engine import CalcJob
from .process import AiidaProcessDocumenter, AiidaProcessDirective


def setup_extension(app):
    app.add_directive_to_domain('py', AiidaCalcJobDocumenter.directivetype, AiidaCalcJobDirective)
    # app.add_autodocumenter(AiidaCalcJobDocumenter)


class AiidaCalcJobDocumenter(AiidaProcessDocumenter):
    """Sphinx Documenter for AiiDA CalcJobs."""
    directivetype = 'aiida-calcjob'
    objtype = 'calcjob'
    priority = 20

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        return inspect.isclass(member) and issubclass(member, CalcJob)


class AiidaCalcJobDirective(AiidaProcessDirective):
    signature = 'CalcJob'
    annotation = 'calcjob'



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=global-statement
"""Definition of AiiDA's process persister and the necessary object loaders."""

import importlib
import logging
import traceback

import plumpy

from aiida.orm.utils import serialize

__all__ = ('AiiDAPersister', 'ObjectLoader', 'get_object_loader')

LOGGER = logging.getLogger(__name__)
OBJECT_LOADER = None


class ObjectLoader(plumpy.DefaultObjectLoader):
    """Custom object loader for `aiida-core`."""

    def load_object(self, identifier):
        """Attempt to load the object identified by the given `identifier`.

        .. note:: We override the `plumpy.DefaultObjectLoader` to be able to throw an `ImportError` instead of a
            `ValueError` which in the context of `aiida-core` is not as apt, since we are loading classes.

        :param identifier: concatenation of module and resource name
        :return: loaded object
        :raises ImportError: if the object cannot be loaded
        """
        module, name = identifier.split(':')
        try:
            module = importlib.import_module(module)
        except ImportError:
            raise ImportError(f"module '{module}' from identifier '{identifier}' could not be loaded")

        try:
            return getattr(module, name)
        except AttributeError:
            raise ImportError(f"object '{name}' from identifier '{identifier}' could not be loaded")


def get_object_loader():
    """Return the global AiiDA object loader.

    :return: The global object loader
    :rtype: :class:`plumpy.ObjectLoader`
    """
    global OBJECT_LOADER
    if OBJECT_LOADER is None:
        OBJECT_LOADER = ObjectLoader()
    return OBJECT_LOADER


class AiiDAPersister(plumpy.Persister):
    """Persister to take saved process instance states and persisting them to the database."""

    def save_checkpoint(self, process, tag=None):
        """Persist a Process instance.

        :param process: :class:`aiida.engine.Process`
        :param tag: optional checkpoint identifier to allow distinguishing multiple checkpoints for the same process
        :raises: :class:`plumpy.PersistenceError` Raised if there was a problem saving the checkpoint
        """
        LOGGER.debug('Persisting process<%d>', process.pid)

        if tag is not None:
            raise NotImplementedError('Checkpoint tags not supported yet')

        try:
            bundle = plumpy.Bundle(process, plumpy.LoadSaveContext(loader=get_object_loader()))
        except ImportError:
            # Couldn't create the bundle
            raise plumpy.PersistenceError(f"Failed to create a bundle for '{process}': {traceback.format_exc()}")

        try:
            process.node.set_checkpoint(serialize.serialize(bundle))
        except Exception:
            raise plumpy.PersistenceError(f"Failed to store a checkpoint for '{process}': {traceback.format_exc()}")

        return bundle

    def load_checkpoint(self, pid, tag=None):
        """Load a process from a persisted checkpoint by its process id.

        :param pid: the process id of the :class:`plumpy.Process`
        :param tag: optional checkpoint identifier to allow retrieving a specific sub checkpoint
        :return: a bundle with the process state
        :rtype: :class:`plumpy.Bundle`
        :raises: :class:`plumpy.PersistenceError` Raised if there was a problem loading the checkpoint
        """
        from aiida.common.exceptions import MultipleObjectsError, NotExistent
        from aiida.orm import load_node

        if tag is not None:
            raise NotImplementedError('Checkpoint tags not supported yet')

        try:
            calculation = load_node(pid)
        except (MultipleObjectsError, NotExistent):
            raise plumpy.PersistenceError(f'Failed to load the node for process<{pid}>: {traceback.format_exc()}')

        checkpoint = calculation.checkpoint

        if checkpoint is None:
            raise plumpy.PersistenceError(f'Calculation<{calculation.pk}> does not have a saved checkpoint')

        try:
            bundle = serialize.deserialize(checkpoint)
        except Exception:
            raise plumpy.PersistenceError(f'Failed to load the checkpoint for process<{pid}>: {traceback.format_exc()}')

        return bundle

    def get_checkpoints(self):
        """Return a list of all the current persisted process checkpoints

        :return: list of PersistedCheckpoint tuples with element containing the process id and optional checkpoint tag.
        """

    def get_process_checkpoints(self, pid):
        """Return a list of all the current persisted process checkpoints for the specified process.

        :param pid: the process pid
        :return: list of PersistedCheckpoint tuples with element containing the process id and optional checkpoint tag.
        """

    def delete_checkpoint(self, pid, tag=None):
        """Delete a persisted process checkpoint, where no error will be raised if the checkpoint does not exist.

        :param pid: the process id of the :class:`plumpy.Process`
        :param tag: optional checkpoint identifier to allow retrieving a specific sub checkpoint
        """
        from aiida.orm import load_node

        calc = load_node(pid)
        calc.delete_checkpoint()

    def delete_process_checkpoints(self, pid):
        """Delete all persisted checkpoints related to the given process id.

        :param pid: the process id of the :class:`aiida.engine.processes.process.Process`
        """



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Data plugin to model an atomic orbital."""
import copy

from aiida.common.exceptions import ValidationError, InputValidationError
from aiida.plugins import OrbitalFactory
from .data import Data

__all__ = ('OrbitalData',)


class OrbitalData(Data):
    """
    Used for storing collections of orbitals, as well as
    providing methods for accessing them internally.
    """

    def clear_orbitals(self):
        """
        Remove all orbitals that were added to the class
        Cannot work if OrbitalData has been already stored
        """
        self.set_attribute('orbital_dicts', [])

    def get_orbitals(self, **kwargs):
        """
        Returns all orbitals by default. If a site is provided, returns
        all orbitals cooresponding to the location of that site, additional
        arguments may be provided, which act as filters on the retrieved
        orbitals.

        :param site: if provided, returns all orbitals with position of site
        :kwargs: attributes than can filter the set of returned orbitals
        :return list_of_outputs: a list of orbitals
        """

        orbital_dicts = copy.deepcopy(self.get_attribute('orbital_dicts', None))
        if orbital_dicts is None:
            raise AttributeError('Orbitals must be set before being retrieved')

        filter_dict = {}
        filter_dict.update(kwargs)
        # prevents KeyError from occuring
        orbital_dicts = [x for x in orbital_dicts if all([y in x for y in filter_dict])]
        orbital_dicts = [x for x in orbital_dicts if all([x[y] == filter_dict[y] for y in filter_dict])]

        list_of_outputs = []
        for orbital_dict in orbital_dicts:
            try:
                orbital_type = orbital_dict.pop('_orbital_type')
            except KeyError:
                raise ValidationError(f'No _orbital_type found in: {orbital_dict}')

            cls = OrbitalFactory(orbital_type)
            orbital = cls(**orbital_dict)
            list_of_outputs.append(orbital)
        return list_of_outputs

    def set_orbitals(self, orbitals):
        """
        Sets the orbitals into the database. Uses the orbital's inherent
        set_orbital_dict method to generate a orbital dict string.

        :param orbital: an orbital or list of orbitals to be set
        """
        if not isinstance(orbitals, list):
            orbitals = [orbitals]
        orbital_dicts = []

        for orbital in orbitals:
            orbital_dict = copy.deepcopy(orbital.get_orbital_dict())
            try:
                _orbital_type = orbital_dict['_orbital_type']
            except KeyError:
                raise InputValidationError(f'No _orbital_type found in: {orbital_dict}')
            orbital_dicts.append(orbital_dict)
        self.set_attribute('orbital_dicts', orbital_dicts)


##########################################################################
#     Here are some ideas for potential future convenience methods
#########################################################################
#     def set_projection_on_site(self, orbital, site, tag=None):
#         """
#         Sets a orbital on a site
#         We prepare the description dictionary, using information `parsed`
#         from the site.
#         """
#         diffusivity = from_site_guess_diffusivity(site) # or 1.
#         position = site.position
#         description = {'somedictionary of the above':''}
#         self.set_projection(orbital=orbital, description=description)

#     def delete_projections_by_attribute(self, selection_attributes):
#         """
#         Deletes all projections whose internal attributes correspond to the
#         selection_attributes
#         """
#         raise NotImplementedError
#
#     def modify_projections(self, key_attributes_to_select_projections, attributes_to_be_modified):
#         """
#         Modifies the projections, as selected by the key_attributes.
#         Overwrites attributes inside these projections, to values stored
#         in attributes_to_be_modified
#         """
#
#    def set_realhydrogenorbitals_from_structure(self, structure, pseudo_family=None):
#        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Simple wrapper around the alembic command line tool that first loads an AiiDA profile."""

import alembic
import click

from aiida.cmdline.params import options


def execute_alembic_command(command_name, **kwargs):
    """Execute an Alembic CLI command.

    :param command_name: the sub command name
    :param kwargs: parameters to pass to the command
    """
    from aiida.backends.sqlalchemy.manager import SqlaBackendManager

    manager = SqlaBackendManager()

    with manager.alembic_config() as config:
        command = getattr(alembic.command, command_name)
        command(config, **kwargs)


@click.group()
@options.PROFILE(required=True)
def alembic_cli(profile):
    """Simple wrapper around the alembic command line tool that first loads an AiiDA profile."""
    from aiida.manage.configuration import load_profile
    from aiida.manage.manager import get_manager

    load_profile(profile=profile.name)
    manager = get_manager()
    manager._load_backend(schema_check=False)  # pylint: disable=protected-access


@alembic_cli.command('revision')
@click.argument('message')
def alembic_revision(message):
    """Create a new database revision."""
    execute_alembic_command('revision', message=message, autogenerate=True)


@alembic_cli.command('current')
@options.VERBOSE()
def alembic_current(verbose):
    """Show the current revision."""
    execute_alembic_command('current', verbose=verbose)


@alembic_cli.command('history')
@click.option('-r', '--rev-range')
@options.VERBOSE()
def alembic_history(rev_range, verbose):
    """Show the history for the given revision range."""
    execute_alembic_command('history', rev_range=rev_range, verbose=verbose)


@alembic_cli.command('upgrade')
@click.argument('revision', type=click.STRING)
def alembic_upgrade(revision):
    """Upgrade the database to the given REVISION."""
    execute_alembic_command('upgrade', revision=revision)


@alembic_cli.command('downgrade')
@click.argument('revision', type=click.STRING)
def alembic_downgrade(revision):
    """Downgrade the database to the given REVISION."""
    execute_alembic_command('downgrade', revision=revision)


if __name__ == '__main__':
    alembic_cli()  # pylint: disable=no-value-for-parameter



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Provides all parameter types."""

from .calculation import CalculationParamType
from .choice import LazyChoice
from .code import CodeParamType
from .computer import ComputerParamType, ShebangParamType, MpirunCommandParamType
from .config import ConfigOptionParamType
from .data import DataParamType
from .group import GroupParamType
from .identifier import IdentifierParamType
from .multiple import MultipleValueParamType
from .node import NodeParamType
from .process import ProcessParamType
from .strings import (NonEmptyStringParamType, EmailType, HostnameType, EntryPointType, LabelStringType)
from .path import AbsolutePathParamType, PathOrUrl, FileOrUrl
from .plugin import PluginParamType
from .profile import ProfileParamType
from .user import UserParamType
from .test_module import TestModuleParamType
from .workflow import WorkflowParamType

__all__ = (
    'LazyChoice', 'IdentifierParamType', 'CalculationParamType', 'CodeParamType', 'ComputerParamType',
    'ConfigOptionParamType', 'DataParamType', 'GroupParamType', 'NodeParamType', 'MpirunCommandParamType',
    'MultipleValueParamType', 'NonEmptyStringParamType', 'PluginParamType', 'AbsolutePathParamType', 'ShebangParamType',
    'UserParamType', 'TestModuleParamType', 'ProfileParamType', 'WorkflowParamType', 'ProcessParamType', 'PathOrUrl',
    'FileOrUrl'
)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name
"""Data migration for legacy `JobCalculations`.

These old nodes have already been migrated to the correct `CalcJobNode` type in a previous migration, but they can
still contain a `state` attribute with a deprecated `JobCalcState` value and they are missing a value for the
`process_state`, `process_status`, `process_label` and `exit_status`. The `process_label` is impossible to infer
consistently in SQL so it will be omitted. The other  will be mapped from the `state` attribute as follows:

.. code-block:: text

    Old state            | Process state  | Exit status | Process status
    ---------------------|----------------|-------------|----------------------------------------------------------
    `NEW`                | `killed`       |    `None`   | Legacy `JobCalculation` with state `NEW`
    `TOSUBMIT`           | `killed`       |    `None`   | Legacy `JobCalculation` with state `TOSUBMIT`
    `SUBMITTING`         | `killed`       |    `None`   | Legacy `JobCalculation` with state `SUBMITTING`
    `WITHSCHEDULER`      | `killed`       |    `None`   | Legacy `JobCalculation` with state `WITHSCHEDULER`
    `COMPUTED`           | `killed`       |    `None`   | Legacy `JobCalculation` with state `COMPUTED`
    `RETRIEVING`         | `killed`       |    `None`   | Legacy `JobCalculation` with state `RETRIEVING`
    `PARSING`            | `killed`       |    `None`   | Legacy `JobCalculation` with state `PARSING`
    `SUBMISSIONFAILED`   | `excepted`     |    `None`   | Legacy `JobCalculation` with state `SUBMISSIONFAILED`
    `RETRIEVALFAILED`    | `excepted`     |    `None`   | Legacy `JobCalculation` with state `RETRIEVALFAILED`
    `PARSINGFAILED`      | `excepted`     |    `None`   | Legacy `JobCalculation` with state `PARSINGFAILED`
    `FAILED`             | `finished`     |      2      |  -
    `FINISHED`           | `finished`     |      0      |  -
    `IMPORTED`           |       -        |      -      |  -

Note the `IMPORTED` state was never actually stored in the `state` attribute, so we do not have to consider it.
The old `state` attribute has to be removed after the data is migrated, because its value is no longer valid or useful.

Note: in addition to the three attributes mentioned in the table, all matched nodes will get `Legacy JobCalculation` as
their `process_label` which is one of the default columns of `verdi process list`.
"""

# Remove when https://github.com/PyCQA/pylint/issues/1931 is fixed
# pylint: disable=no-name-in-module,import-error
from django.db import migrations
from aiida.backends.djsite.db.migrations import upgrade_schema_version

REVISION = '1.0.38'
DOWN_REVISION = '1.0.37'


class Migration(migrations.Migration):
    """Data migration for legacy `JobCalculations`."""

    dependencies = [
        ('db', '0037_attributes_extras_settings_json'),
    ]

    # Note that the condition on matching target nodes is done only on the `node_type` amd the `state` attribute value.
    # New `CalcJobs` will have the same node type and while their active can have a `state` attribute with a value
    # of the enum `CalcJobState`, some of which match the deprecated `JobCalcState`, however, the new ones are stored
    # in lower case, so we do not run the risk of matching them by accident.
    operations = [
        migrations.RunSQL(
            sql=r"""
                UPDATE db_dbnode
                SET attributes = attributes - 'state' || '{"process_state": "killed", "process_status": "Legacy `JobCalculation` with state `NEW`", "process_label": "Legacy JobCalculation"}'
                WHERE node_type = 'process.calculation.calcjob.CalcJobNode.' AND attributes @> '{"state": "NEW"}';
                UPDATE db_dbnode
                SET attributes = attributes - 'state' || '{"process_state": "killed", "process_status": "Legacy `JobCalculation` with state `TOSUBMIT`", "process_label": "Legacy JobCalculation"}'
                WHERE node_type = 'process.calculation.calcjob.CalcJobNode.' AND attributes @> '{"state": "TOSUBMIT"}';
                UPDATE db_dbnode
                SET attributes = attributes - 'state' || '{"process_state": "killed", "process_status": "Legacy `JobCalculation` with state `SUBMITTING`", "process_label": "Legacy JobCalculation"}'
                WHERE node_type = 'process.calculation.calcjob.CalcJobNode.' AND attributes @> '{"state": "SUBMITTING"}';
                UPDATE db_dbnode
                SET attributes = attributes - 'state' || '{"process_state": "killed", "process_status": "Legacy `JobCalculation` with state `WITHSCHEDULER`", "process_label": "Legacy JobCalculation"}'
                WHERE node_type = 'process.calculation.calcjob.CalcJobNode.' AND attributes @> '{"state": "WITHSCHEDULER"}';
                UPDATE db_dbnode
                SET attributes = attributes - 'state' || '{"process_state": "killed", "process_status": "Legacy `JobCalculation` with state `COMPUTED`", "process_label": "Legacy JobCalculation"}'
                WHERE node_type = 'process.calculation.calcjob.CalcJobNode.' AND attributes @> '{"state": "COMPUTED"}';
                UPDATE db_dbnode
                SET attributes = attributes - 'state' || '{"process_state": "killed", "process_status": "Legacy `JobCalculation` with state `RETRIEVING`", "process_label": "Legacy JobCalculation"}'
                WHERE node_type = 'process.calculation.calcjob.CalcJobNode.' AND attributes @> '{"state": "RETRIEVING"}';
                UPDATE db_dbnode
                SET attributes = attributes - 'state' || '{"process_state": "killed", "process_status": "Legacy `JobCalculation` with state `PARSING`", "process_label": "Legacy JobCalculation"}'
                WHERE node_type = 'process.calculation.calcjob.CalcJobNode.' AND attributes @> '{"state": "PARSING"}';
                UPDATE db_dbnode
                SET attributes = attributes - 'state' || '{"process_state": "excepted", "process_status": "Legacy `JobCalculation` with state `SUBMISSIONFAILED`", "process_label": "Legacy JobCalculation"}'
                WHERE node_type = 'process.calculation.calcjob.CalcJobNode.' AND attributes @> '{"state": "SUBMISSIONFAILED"}';
                UPDATE db_dbnode
                SET attributes = attributes - 'state' || '{"process_state": "excepted", "process_status": "Legacy `JobCalculation` with state `RETRIEVALFAILED`", "process_label": "Legacy JobCalculation"}'
                WHERE node_type = 'process.calculation.calcjob.CalcJobNode.' AND attributes @> '{"state": "RETRIEVALFAILED"}';
                UPDATE db_dbnode
                SET attributes = attributes - 'state' || '{"process_state": "excepted", "process_status": "Legacy `JobCalculation` with state `PARSINGFAILED`", "process_label": "Legacy JobCalculation"}'
                WHERE node_type = 'process.calculation.calcjob.CalcJobNode.' AND attributes @> '{"state": "PARSINGFAILED"}';
                UPDATE db_dbnode
                SET attributes = attributes - 'state' || '{"process_state": "finished", "exit_status": 2, "process_label": "Legacy JobCalculation"}'
                WHERE node_type = 'process.calculation.calcjob.CalcJobNode.' AND attributes @> '{"state": "FAILED"}';
                UPDATE db_dbnode
                SET attributes = attributes - 'state' || '{"process_state": "finished", "exit_status": 0, "process_label": "Legacy JobCalculation"}'
                WHERE node_type = 'process.calculation.calcjob.CalcJobNode.' AND attributes @> '{"state": "FINISHED"}';
                """,
            reverse_sql=''
        ),
        upgrade_schema_version(REVISION, DOWN_REVISION)
    ]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=import-error,no-name-in-module
"""Module to manage node settings for the SQLA backend."""
from pytz import UTC

from sqlalchemy import Column
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy.types import Integer, String, DateTime

from aiida.backends import sqlalchemy as sa
from aiida.backends.sqlalchemy.models.base import Base
from aiida.common import timezone


class DbSetting(Base):
    """Class to store node settings using the SQLA backend."""
    __tablename__ = 'db_dbsetting'
    __table_args__ = (UniqueConstraint('key'),)
    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name

    key = Column(String(255), index=True, nullable=False)
    val = Column(JSONB, default={})

    # I also add a description field for the variables
    description = Column(String(255), default='', nullable=False)
    time = Column(DateTime(timezone=True), default=UTC, onupdate=timezone.now)

    def __str__(self):
        return f"'{self.key}'={self.getvalue()}"

    @classmethod
    def set_value(cls, key, value, other_attribs=None, stop_if_existing=False):
        """Set a setting value."""
        other_attribs = other_attribs if other_attribs is not None else {}
        setting = sa.get_scoped_session().query(DbSetting).filter_by(key=key).first()
        if setting is not None:
            if stop_if_existing:
                return
        else:
            setting = cls()

        setting.key = key
        setting.val = value
        flag_modified(setting, 'val')
        setting.time = timezone.datetime.now(tz=UTC)
        if 'description' in other_attribs.keys():
            setting.description = other_attribs['description']
        setting.save()

    def getvalue(self):
        """This can be called on a given row and will get the corresponding value."""
        return self.val

    def get_description(self):
        """This can be called on a given row and will get the corresponding description."""
        return self.description

    @classmethod
    def del_value(cls, key):
        """Delete a setting value."""
        setting = sa.get_scoped_session().query(DbSetting).filter(key=key)
        setting.val = None
        setting.time = timezone.datetime.utcnow()
        flag_modified(setting, 'val')
        setting.save()
